#include "StO97DstMaker.h"


ClassImp(StO97DstMaker)


//
// Set maximum file size to 1.9 GB (Root has a 2GB limit)
//
#define MAXFILESIZE 1900000000


//_________________
StO97DstMaker::StO97DstMaker(const Char_t *iFileName,
                             const Char_t *oFileName) {
    std::cout << "[StO97DstMaker] Creating an instance... ";
    TDatime d;

    unsigned int num;
    TString ifname(iFileName);
    int dotPos = ifname.Last('.');
    ifname[dotPos] = '\0';
    num = atoi(ifname.Data() + ifname.Last('_') + 1);
    ifname[dotPos] = '.';
    mOutFileName = Form("%s_%04d%02d%02d_%d.root", oFileName,
                        d.GetYear(), d.GetMonth(), d.GetDay(), num);
    std::cout << "[StO97DstMaker]: output file name " << mOutFileName << std::endl;

    mInFileName  = iFileName;
    mTree = 0;
    mOutFile = 0;
    mInFile = 0;
    mCompression = 9;
    mNEvents = 0;
    mStop = false;

    matrix  = new TMatrixTSym<double>(2); // matrix for calculation of transverse sphericity
    matrix2 = new TMatrixTSym<double>(2); // matrix for calculation of transverse sphericity 2

    std::cout << "[OK]" << std::endl;
}


//_________________
StO97DstMaker::~StO97DstMaker() {
    /* nothing to do */
}


//_________________
Int_t StO97DstMaker::Init() {
    std::cout << "[StO97DstMaker] ---=== Initializing maker ===---" << std::endl;
    mEvent = new StO97Event();
    //
    // Input file
    //
    std::cout << "[StO97DstMaker] Openning an input file... ";
    if (mInFile) {
        fclose(mInFile);
    }
    mInFile = fopen(mInFileName, "r");
    if (mInFile) {
        std::cout << "[OK]" << std::endl;
    }
    else {
        std::cout << "[FAIL]" << std::endl;
        return kStFatal;
    }
    //
    // Output file
    //
    std::cout << "[StO97DstMaker] Creating output file... ";
    if (mOutFile) {
        mOutFile->Close();
        mOutFile->Open(mOutFileName, "RECREATE");
    }
    else {
        mOutFile = new TFile(mOutFileName, "RECREATE");
    }
    if (mOutFile) {
        std::cout << "[OK]" << std::endl;
    }
    else {
        std::cout << "[FAIL]" << std::endl;
        return kStFatal;
    }
    SetCompressionlevel(mCompression); // default value from constructor
    //
    // TTree setup
    //
    std::cout << "[StO97DstMaker] Creating TTree... ";
    if (mTree) {
        delete mTree;
    }
    mTree = new TTree("StO97Dst", "StO97Dst");
    if (!mTree) {
        std::cout << "[FAIL]" << std::endl;
        return kStFatal;
    }
    mTree->SetMaxTreeSize(MAXFILESIZE);
    mTree->Branch("StO97Event", "StO97Event", &mEvent);
    mNBytes = 0;
    std::cout << "[OK]" << std::endl;
    //
    // Finalization of initialization :)
    //
    std::cout << "[StO97DstMaker] ---=== Initialization complite ===---" << std::endl;

    return StMaker::Init();
}


//_________________
void StO97DstMaker::SetCompressionlevel(Int_t comp) {
    mCompression = comp;
    std::cout << "[StO97DstMaker] Compression level = " << mCompression << std::endl;
    mOutFile->SetCompressionLevel(mCompression);
}


//_________________
Int_t StO97DstMaker::Make() {
    //
    // Read event
    //
    mNEvents++;
    mStop = !mStop;
    mStop = ReadEvent();
    if (mStop) {
        return kStOk;
    }
    //
    // Read tracks
    //
    mStop = ReadTracks();
    if (!mStop)
    {
        mEvent->SetTransverseSphericity(mSperp);
        mEvent->SetTransverseSphericity2(mSperp2);
        mNBytes += mTree->Fill();
    }

    return kStOk;
}


//_________________
Int_t StO97DstMaker::Finish() {
    std::cout << "[StO97DstMaker] ---=== Finalization of maker ===---"
        << "[StO97DstMaker] total number of events = "
        << mNEvents << std::endl
        << "[StO97DstMaker] total bytes = "
        << mNBytes
        << std::endl;
    if (mInFile) {
        fclose(mInFile);
        mInFile = 0;
    }
    if (mTree) {
        if (mOutFile) {
            mOutFile->Write();
        }
        delete mTree;
        mTree = 0;
    }
    if (mOutFile) {
        mOutFile->Close();
        delete mOutFile;
        mOutFile = 0;
    }

    return kStOk;
}


//_________________
Bool_t StO97DstMaker::ReadEvent() {
    Int_t eventNumber;
    Float_t impPar;
    Float_t rot;
    Char_t buf[128];

    while (1) {
        Int_t ret = fscanf(mInFile, " %i %i %f %f",
                           &eventNumber, &mNTracks, &impPar, &rot);
        if (ret <= 0) {               // if can't read a line
            fgets(buf, 127, mInFile); // skip it
        }
        else if (ret == 4) { // 4 - total number of event parameters
            mEvent->Clear();
            mEvent->SetEventNumber(eventNumber);
            mEvent->SetImpactPar(impPar);
            mEvent->SetEventPlaneRot(rot);
            return false;
        }
        if (feof(mInFile)) { // if EOF then return stop flag
            return true;
        }
    }
}

//_________________
Bool_t StO97DstMaker::ReadTracks() {
    Int_t id;
    Int_t pdgId;
    Float_t px, py, pz;
    Float_t energy, mass;
    Float_t xFr, yFr, zFr, tFr;
    TLorentzVector particle;

    Float_t pt_full  = 0.0;
    Float_t pt_full2 = 0.0;
    mSperp  = -999.;
    mSperp2 = -999.;

    matrix->Zero();
    matrix2->Zero();

    for (Int_t iTrack = 0; iTrack < mNTracks; iTrack++) {
        Int_t ret = fscanf(mInFile, " %i %i %f %f %f %f %f %f %f %f %f",
                           &id, &pdgId, &px, &py, &pz,
                           &energy, &mass, &xFr, &yFr, &zFr, &tFr);
        if (ret != 11) {
            return true;
        }
        else {
            StO97Track *track = mEvent->AddOscar97Track();
            track->SetId(id);
            track->SetPdgId(pdgId);
            track->SetPx(px);
            track->SetPy(py);
            track->SetPz(pz);
            track->SetMass(mass);
            track->SetXfr(xFr);
            track->SetYfr(yFr);
            track->SetZfr(zFr);
            track->SetTfr(tFr);

            /*
             * Transverse sphericity calculation with following cuts:
             *   * |eta| < 0.5 && |eta| < 1.0
             *   * pT > 0.15 GeV/c
             */
            // setup particle vector
            particle.SetPxPyPzE(px, py, pz, energy);
            Float_t pt  = particle.Pt();

            if (pt > 0.15) {
                Float_t eta = particle.PseudoRapidity(); // it heals pt = 0 case
                if (TMath::Abs(eta) < 0.5)
                {
                    (*matrix)(0, 0) += px*px/pt;
                    (*matrix)(1, 1) += py*py/pt;
                    (*matrix)(0, 1) += px*py/pt;
                    (*matrix)(1, 0) += px*py/pt;

                    pt_full += pt;
                }
                if (TMath::Abs(eta) < 1.0)
                {
                    (*matrix2)(0, 0) += px*px/pt;
                    (*matrix2)(1, 1) += py*py/pt;
                    (*matrix2)(0, 1) += px*py/pt;
                    (*matrix2)(1, 0) += px*py/pt;

                    pt_full2 += pt;
                }
            }
        }
    }

    if (pt_full != 0.0) {
        // |eta| < 0.5 case
        *matrix  *= 1./pt_full;
        TMatrixDSymEigen death_machine(*matrix);
        TVectorD eigen = death_machine.GetEigenValues();
        mSperp = 2.*eigen.Min()/(eigen[0] + eigen[1]);

        // |eta| < 1.0 case
        *matrix2 *= 1./pt_full2;
        TMatrixDSymEigen death_machine2(*matrix2);
        TVectorD eigen2 = death_machine2.GetEigenValues();
        mSperp2 = 2.*eigen2.Min()/(eigen2[0] + eigen2[1]);
    }
    return false;
}
