/***************************************************************************
 *
 * $Id: QinvCorrFctnKt_th.cxx,v 1.3 2003/09/02 17:58:20 perev Exp $
 *
 * Author: Mercedes Lopez Noriega, OSU, mercedes@pacific.mps.ohio-state.edu
 ***************************************************************************
 *
 * Description: part of STAR HBT Framework: StHbtMaker package
 *   QinvCFs for a given number of CFs (nCFs) between ktLo and ktHi 
 *   where kt is the four-momentum of the pair 
 *
 ***************************************************************************
 *
 * $Log: QinvCorrFctnKt_th.cxx,v $
 * Revision 1.3  2003/09/02 17:58:20  perev
 * gcc 3.2 updates + WarnOff
 *
 * Revision 1.2  2003/01/31 19:21:09  magestro
 * Cleared up simple compiler warnings on i386_linux24
 *
 * Revision 1.1  2002/05/17 14:26:21  mercedes
 * N Qinv CFs (kt bins) between ktmin and ktmax
 *
 *
 **************************************************************************/

#include "StHbtMaker/CorrFctn/QinvCorrFctnKt_th.h"
//#include "StHbtMaker/Infrastructure/StHbtHisto.hh"
#include <cstdio>
#include <sstream>

#ifdef __ROOT__ 
ClassImp(QinvCorrFctnKt_th)
#endif

//____________________________
QinvCorrFctnKt_th::QinvCorrFctnKt_th(char* title,
                                     const int& nbins, const float& QinvLo,
                                     const float& QinvHi,
                                     const int& nCFs, const float& KtLo,
                                     const float& KtHi)
{
  mNumberCFs = nCFs;
  mKtMin = KtLo;
  mKtMax = KtHi;
  mDeltaKt = (mKtMax-mKtMin)/mNumberCFs;

  // set up numerators
  mNumerator = new StHbt1DHisto[mNumberCFs];

  // set up denominators
  mDenominator = new StHbt1DHisto[mNumberCFs];

  stringstream TitNum,TitDen,TitCurrent;
  TitNum.str("");TitDen.str("");
  TitNum<<title<<"_Num";
  TitDen<<title<<"_Den";

  for (int i=0; i<mNumberCFs; i++)
  {
    TitCurrent.str("");
    TitCurrent<<TitNum.str().c_str()<<"_bin_"<<i;
    mNumerator[i].SetName(TitCurrent.str().c_str());
    mNumerator[i].SetTitle(TitCurrent.str().c_str());
    mNumerator[i].SetBins(nbins,QinvLo,QinvHi);
    mNumerator[i].SetDirectory(0);
    mNumerator[i].Sumw2();

    TitCurrent.str("");
    TitCurrent<<TitDen.str().c_str()<<"_bin_"<<i;
    mDenominator[i].SetName(TitCurrent.str().c_str());
    mDenominator[i].SetTitle(TitCurrent.str().c_str());
    mDenominator[i].SetBins(nbins,QinvLo,QinvHi);
    mDenominator[i].SetDirectory(0);
    mDenominator[i].Sumw2();
  }
}

//_________________
QinvCorrFctnKt_th::~QinvCorrFctnKt_th()
{
  delete [] mNumerator;
  delete [] mDenominator;
}

//_________________
void QinvCorrFctnKt_th::WriteOutHistos()
{

  for(int i=0; i<mNumberCFs; i++)
  {
    mNumerator[i].Write();
    mDenominator[i].Write();
  }
}

//_________________
void QinvCorrFctnKt_th::Finish()
{
  /* Ratio creation might be here */
}

//_________________
StHbtString QinvCorrFctnKt_th::Report()
{
  int mNumeratorEntries=0;
  int mDenominatorEntries=0;

  for (int i=0; i<mNumberCFs; i++)
  {
    mNumeratorEntries += (int)mNumerator[i].GetEntries();
    mDenominatorEntries += (int)mDenominator[i].GetEntries();
  }
  string stemp = "Qinv Correlation Function Report:\n";
  char ctemp[100];
  sprintf(ctemp,"Number of entries in all numerators:\t%i\n",mNumeratorEntries);
  stemp += ctemp;
  sprintf(ctemp,"Number of entries in all denominators:\t%i\n",mDenominatorEntries);
  stemp += ctemp;
  StHbtString returnThis = stemp;
  return returnThis;
}

//____________________________
void QinvCorrFctnKt_th::AddRealPair(const StHbtPair* pair)
{
  int mIndex = (int)(((fabs(pair->kT())-mKtMin)/mDeltaKt));
  if ((mIndex>=0)&&(mIndex<mNumberCFs))
  {
    // get a pair of tracks from StHbtPair
    StHbtTrack *track1 = pair->track1()->Track();
    StHbtTrack *track2 = pair->track2()->Track();

    // get freeze-out components or in case of not sim track just leave it alone
    if (track1->IsSimTrack() && track2->IsSimTrack())
    {
      double hc = 0.1973269718; // GeV*fm
      double hchalf = hc*0.5;
          
      // first track
      float xfr1 = track1->FreezeOutX();
      float yfr1 = track1->FreezeOutY();
      float zfr1 = track1->FreezeOutZ();
      float tfr1 = track1->FreezeOutT();
      float E1   = sqrt(track1->Ptot2() + track1->TofMassSqr());
      float px1  = track1->P().x();
      float py1  = track1->P().y();
      float pz1  = track1->P().z();

      // second track
      float xfr2 = track2->FreezeOutX();
      float yfr2 = track2->FreezeOutY();
      float zfr2 = track2->FreezeOutZ();
      float tfr2 = track2->FreezeOutT();
      float E2   = sqrt(track2->Ptot2() + track2->TofMassSqr());
      float px2  = track2->P().x();
      float py2  = track2->P().y();
      float pz2  = track2->P().z();

      // check uncertinity principle
      if (fabs(xfr1 - xfr2)*fabs(px1 - px2) < hchalf ||
          fabs(yfr1 - yfr2)*fabs(py1 - py2) < hchalf ||
          fabs(zfr1 - zfr2)*fabs(pz1 - pz2) < hchalf ||
          fabs(E1 - E2)*fabs(tfr1 - tfr2) < hchalf)       return;

      // create two TLorentzVectors and calculate weight of two particles
      TLorentzVector X1(xfr1, yfr1, zfr1, tfr1);
      TLorentzVector X2(xfr2, yfr2, zfr2, tfr2);

      TLorentzVector P1(px1, py1, pz1, E1);
      TLorentzVector P2(px2, py2, pz2, E2);

      TLorentzVector deltaX = X1 - X2;
      TLorentzVector deltaP = P1 - P2;

      double weight = 1.0 + cos(deltaP*deltaX/hc);

      // fill numerator with qinv and weight
      mNumerator[mIndex].Fill(fabs(pair->qInv()), weight);
      // fill denominator with 1.0 weight
      mDenominator[mIndex].Fill(fabs(pair->qInv()), 1.0);
    }
  }
}

//____________________________
void QinvCorrFctnKt_th::AddMixedPair(const StHbtPair* pair)
{
  /* NO-OP */
}


