#include "StSphericity.h"
#include "StThreeVector.hh"
#include "StThreeVectorF.hh"
#include "StPhysicalHelixD.hh"

ClassImp(StSphericity)



StSphericity::StSphericity(StMuDstMaker *muMaker,
                           const Char_t *oFileName)
{
    mMuDstMaker  = muMaker;
    mOutFileName = oFileName;
    mNEventsIn   = 0;
    ptCut        = 0.5;
    matrix       = new TMatrixTSym<double>(2);
    matrix3d     = new TMatrixTSym<double>(3);
    sphLoCut     = 0.1;
    sphHiCut     = 0.5;
}



StSphericity::~StSphericity()
{
    delete matrix;
}



Int_t StSphericity::Init()
{
    // create an output file
    mOutFile = new TFile(mOutFileName, "RECREATE");
    //
    // Create histograms
    //
    hRefMult = new TH1I("hRefMult", ";RefMult;events", 100, 0, 100);
    // no cut to sphericity
    hEta = new TH1F("hEta", ";#eta;#", 100, -3., 3.);
    hPhi = new TH1F("hPhi", ";#phi [rad];#", 300, -7., 7.);
    hDEtaDPhi = new TH2F("hDEtaDPhi", ";#Delta#eta;#Delta#phi [rad]",
                         200, -3., 3., /* eta */
                         400, -1.5, 6. /* phi */);
    hEtaPhi = new TH2F("hEtaPhi", ";#eta;#phi [rad]", 
                       200, -3., 3., /* eta */
                       400, -1.5, 6. /* phi */);
    // with sphericity cut
    hDEtaDPhi_lo_sphManual = new TH2F("hDEtaDPhi_lo_sphManual", ";#Delta#eta;#Delta#phi [rad]",
                                      200, -3., 3., /* eta */
                                      400, -1.5, 6.  /* phi */);
    hDEtaDPhi_hi_sphManual = new TH2F("hDEtaDPhi_hi_sphManual", ";#Delta#eta;#Delta#phi [rad]",
                                      200, -3., 3., /* eta */
                                      400, -1.5, 6.  /* phi */);
    hDEtaDPhi_lo_sphAuto = new TH2F("hDEtaDPhi_lo_sphAuto", ";#Delta#eta;#Delta#phi [rad]",
                                      200, -3., 3., /* eta */
                                      400, -1.5, 6.  /* phi */);
    hDEtaDPhi_hi_sphAuto = new TH2F("hDEtaDPhi_hi_sphAuto", ";#Delta#eta;#Delta#phi [rad]",
                                      200, -3., 3., /* eta */
                                      400, -1.5, 6.  /* phi */);
    hDEtaDPhi_lo_sph3d = new TH2F("hDEtaDPhi_lo_sph3d", ";#Delta#eta;#Delta#phi [rad]",
                                      200, -3., 3., /* eta */
                                      400, -1.5, 6.  /* phi */);
    hDEtaDPhi_hi_sph3d = new TH2F("hDEtaDPhi_hi_sph3d", ";#Delta#eta;#Delta#phi [rad]",
                                      200, -3., 3., /* eta */
                                      400, -1.5, 6.  /* phi */);
    // transverse and 3d sphericity histograms
    hSphManual = new TH1F("hSphManual", ";Sphericity;#",
                          300, -0.1, 1.1);
    hSphAuto = new TH1F("hSphAuto", ";Sphericity;#",
                        300, -0.1, 1.1);
    hSphDiff = new TH1F("hSphDiff", ";Sphericity_{auto} - Sphericity_{manual};#",
                        200, -1., 1.);
    hSph3dDiff = new TH1F("hSph3dDiff", ";Sphericity_{auto} - Sphericity_{3d};#",
                        200, -1., 1.);
    hSph3d = new TH1F("hSph3d", ";Sphericity_{3D};#",
                      300, -0.1, 1.1);

    return StMaker::Init();
}



void StSphericity::Clear(Option_t *option)
{
    StMaker::Clear();
}



Int_t StSphericity::Finish()
{
    mOutFile->Write();
    mOutFile->Close();

    return kStOk;
}



void StSphericity::AddTrigId(unsigned int aTrigId)
{
    mTrigIdCol.push_back(aTrigId);
}



Int_t StSphericity::Make()
{
    mNEventsIn++;
    mMuDst = NULL;
    mMuEvent = NULL;

    // MuDstMaker initialization
    if(!mMuDstMaker)
    {
        mMuDstMaker = (StMuDstMaker*)GetMaker("MuDst");
        if(!mMuDstMaker)
        {
            cout << "Can not find MuDstMaker\n";
            return kStOk;
        }
    }

    // Obtaining MuDst
    mMuDst = mMuDstMaker->muDst();
    if (!mMuDst)
    {
        mMuDst = (StMuDst*)GetInputDS("MuDst");
        if (!mMuDst)
        {
            cout << "can not obtain MuDst\n";
            return kStOk;
        }
    }

    // Obtaining MuEvent
    mMuEvent = (StMuEvent *)mMuDst->event();


    if (mMuEvent)
    {
        unsigned short refMult = mMuEvent->refMult();
        if (refMult < 0)
        {
            cout << "RefMult less than zero\n";
            return kStOk;
        }

        int nVertices   = mMuDst->numberOfPrimaryVertices();
        int nPrimTracks = mMuDst->numberOfPrimaryTracks();

        if (!nVertices)   return kStOk;

        bool goodTrigger = false;
        if (mTrigIdCol.size() == 0)   goodTrigger = true;
        else
        {
            for (unsigned int iTrg=0; iTrg < mTrigIdCol.size(); iTrg++)
            {
                if (mMuEvent->triggerIdCollection().nominal().isTrigger(mTrigIdCol.at(iTrg)))
                {
                    goodTrigger = true;
                    break;
                }
            }
        }
        if (!goodTrigger)   return kStOk;

        mPrimVertex = mMuDst->primaryVertex(0);
        // pPositive ranking only
        if (mPrimVertex->ranking() <= 0)   return kStOk;
        // not (0,0,0) position of the primary vertex
        if (mPrimVertex->position().x() == 0 &&
            mPrimVertex->position().y() == 0 &&
            mPrimVertex->position().z() == 0)
        {
            cout << "[kStOk] Vertex equals to zero!\n";
            return kStOk;
        }
        // reasonable amount of tracks
        if (nPrimTracks < 0 || nPrimTracks > 10000)
        {
            cout << "[kStOk] N primary tracks cut failed!\n";
            return kStOk;
        }

        StBTofHeader *btofH = mMuDst->btofHeader();
        if (!btofH)
        {
            cout << "StBTofHeader - NULL pointer!\n";
            return kStOk;
        }

        //
        // Calculate the amount of the pile-up vertices
        //
        int mNVtxInBunchCross = 0;
        int nVert = mMuDst->numberOfPrimaryVertices();
        for (int i = 0; i < nVert; i++)
        {
            StMuPrimaryVertex *primVtx = mMuDst->primaryVertex(i);

            if (!primVtx || primVtx->ranking() <= 0.)   continue;

            if (TMath::Abs(primVtx->position().x()) < 1e-5 &&
                TMath::Abs(primVtx->position().y()) < 1e-5 &&
                TMath::Abs(primVtx->position().z()) < 1e-5)      continue;
            if (TMath::Abs(primVtx->position().z() - btofH->vpdVz()) < 5.0)   mNVtxInBunchCross++;
        }
        //
        // There should be only one primary vertex that caused the trigger
        //
        if (mNVtxInBunchCross > 1)   return kStOk;

        int numOfLeadPart = -1;
        float pOfLeadPart = -999.;
        float sphManual   = -999.;
        float sphAuto     = -999.;
        float sph3d       = -999.;
        float pt_full     = 0.0;
        float p_full      = 0.0;
        // (a       m_diag)
        // (m_diag  b     )
        float m_a = 0., m_diag = 0., m_b = 0.;
        float D = 0.0;   // discriminant
        float lambda[2]; // eigenvalues
        unsigned int nPairs = 0;

        // fill matricies with zeros
        matrix->Zero();
        matrix3d->Zero();

        hRefMult->Fill(refMult);

        for (int i = 0; i < nPrimTracks; i++)
        {
            mPrimTrack = mMuDst->primaryTracks(i);
            //
            // Primary track cut's
            //
            if (!AcceptTrack(mPrimTrack))   continue;

            float pt = mPrimTrack->pt();
            const StThreeVectorF &p = mPrimTrack->p();
            if (fabs(mPrimTrack->eta()) < 1.0 && p.mag() > ptCut)
            {
                if (pOfLeadPart < p.mag())
                {
                    numOfLeadPart = i;
                    pOfLeadPart = p.mag();
                }
                float px = p.x();
                float py = p.y();
                float pz = p.z();
                float buf_a = px*px/pt;
                float buf_b = py*py/pt;
                float buf_diag = px*py/pt;

                // fill matrix for 3D sphericity
                //   (row, col)
                (*matrix3d)(0, 0) += px*px;
                (*matrix3d)(0, 1) += px*py;
                (*matrix3d)(0, 2) += px*pz;

                (*matrix3d)(1, 0) += py*px;
                (*matrix3d)(1, 1) += py*py;
                (*matrix3d)(1, 2) += py*pz;

                (*matrix3d)(2, 0) += pz*px;
                (*matrix3d)(2, 1) += pz*py;
                (*matrix3d)(2, 2) += pz*pz;
                // fill matrix for transverse sphericity
                (*matrix)(0, 0) += buf_a;
                (*matrix)(1, 1) += buf_b;
                (*matrix)(0, 1) += buf_diag;
                (*matrix)(1, 0) += buf_diag;
                // manual calcs
                m_a += buf_a;
                m_b += buf_b;
                m_diag += buf_diag;
                pt_full += pt;
                p_full += p.mag();
                nPairs++;
            }
        }

        if (numOfLeadPart < 0 || nPairs < 2)   return kStOk;

        //
        // Normalize Matrices
        //
        m_a /= pt_full;
        m_b /= pt_full;
        m_diag /= pt_full;
        *matrix *= 1./pt_full;
        *matrix3d *= 1./(p_full*p_full);

        //
        // Calculate eigenvalues
        //
        D = pow(m_a + m_b, 2.) - 4.*(m_a*m_b - m_diag*m_diag);
        if (D >= 0.) // two real lambda
        {
            lambda[0] = 0.5*(m_a + m_b + sqrt(D));
            lambda[1] = 0.5*(m_a + m_b - sqrt(D));
            sphManual = 2./(lambda[0] + lambda[1]);
            if (lambda[0] < lambda[1])
                sphManual *= lambda[0];
            else
                sphManual *= lambda[1];
            hSphManual->Fill(sphManual);
        }

        //
        // Automatic calculation of eigenvalues
        //
        // transverse sphericity
        TMatrixDSymEigen death_machine(*matrix);
        TVectorD eigen = death_machine.GetEigenValues();
        sphAuto = 2.*eigen.Min()/(eigen[0] + eigen[1]);
        hSphAuto->Fill(sphAuto);
        hSphDiff->Fill(sphAuto - sphManual);
        // 3D sphericity
        TMatrixDSymEigen death_machine_3d(*matrix3d);
        TVectorD eigen3d = death_machine_3d.GetEigenValues();
        float l1 = eigen3d.Min();
        float l2 = l1;
        float lmax = eigen3d.Max();
        for (int i = 0; i < 3; i++)
        {
            if (eigen3d[i] > l1 && eigen3d[i] < lmax)
            {
                l2 = eigen3d[i];
                break;
            }
        }
        sph3d = 3./2.*(l1 + l2);
        hSph3d->Fill(sph3d);
        hSph3dDiff->Fill(sphAuto - sph3d);

        StMuTrack *leadTrack = mMuDst->primaryTracks(numOfLeadPart);
        for (int i = 0; i < nPrimTracks; i++)
        {
            if (i == numOfLeadPart)   continue;
            StMuTrack *lTrack = mMuDst->primaryTracks(i);
            if (!AcceptTrack(lTrack))   continue;
            if (lTrack->p().mag() > 0.5)
            {
                float phi = lTrack->phi();
                float eta = lTrack->eta();
                float dEta = eta - leadTrack->eta();
                float dPhi = phi - leadTrack->phi();

                hEta->Fill(eta);
                hPhi->Fill(phi);
                hEtaPhi->Fill(eta, phi);
                hDEtaDPhi->Fill(dEta, dPhi);
                // manual calculated sphericity
                if (sphManual >= 0. && sphManual < sphLoCut)
                    hDEtaDPhi_lo_sphManual->Fill(dEta, dPhi);
                if (sphManual > sphHiCut && sphManual <= 1.0)
                    hDEtaDPhi_hi_sphManual->Fill(dEta, dPhi);
                // automatic calculated sphericity
                if (sphAuto >= 0. && sphAuto < sphLoCut)
                    hDEtaDPhi_lo_sphAuto->Fill(dEta, dPhi);
                if (sphAuto > sphHiCut && sphAuto <= 1.0)
                    hDEtaDPhi_hi_sphAuto->Fill(dEta, dPhi);
                // 3d sphericity
                if (sph3d >= 0. && sph3d < sphLoCut)
                    hDEtaDPhi_lo_sph3d->Fill(dEta, dPhi);
                if (sph3d > sphHiCut && sph3d <= 1.0)
                    hDEtaDPhi_hi_sph3d->Fill(dEta, dPhi);
            }
        }
    }
    return kStOk;
}



bool StSphericity::AcceptTrack(StMuTrack *track)
{
    return (track->index2Global() >= 0 &&
            track->vertexIndex() == 0 &&
            track->type() == 1 &&
            track->nHits() >= 15 &&
            track->nHitsFit() >= 15 &&
            track->flag() >= 300 && track->flag() < 1000 &&
            track->dca(0).perp() >= 0. && track->dca(0).perp() <= 3. &&
            track->dcaGlobal(0).perp() >= 0. && track->dcaGlobal(0).perp() <= 3.);
}
