#include "StSphericity.h"
#include "StThreeVector.hh"
#include "StThreeVectorF.hh"
#include "StPhysicalHelixD.hh"

ClassImp(StSphericity)


StSphericity::StSphericity(const char *aDir, 
                           const char *aFileName,
                           const char *aFilter,
                           int aMaxFiles)
{
    dir      = string(aDir);
    fileName = string(aFileName);
    filter   = aFilter;
    maxFiles = aMaxFiles;
    outFileName = "oFemtoQA.root";

    tree       = 0;
    chain      = 0;
    femtoEvent = 0;
    eventId    = 0;
    mTrig      = 0;
}



Int_t StSphericity::Init()
{
    InitRead(dir, fileName, filter, maxFiles);
    nEvents = (unsigned int)chain->GetEntries();
    outFile = new TFile(outFileName, "RECREATE");

    hDEtaDPhi = new TH2F("hDEtaDPhi", ";#Delta#eta;#Delta#phi [rad]", 
                         200, -3., 3., /* eta */
                         400, -10., 10.  /* phi */);
    hEtaPhi = new TH2F("hEtaPhi", ";#eta;#phi [rad]", 
                       200, -3., 3., /* eta */
                       400, -10., 10.  /* phi */);
    hDEtaDPhi_lo_sph = new TH2F("hDEtaDPhi_lo_sph", ";#Delta#eta;#Delta#phi [rad]", 
                                200, -3., 3., /* eta */
                                400, -10., 10.  /* phi */);
    hDEtaDPhi_hi_sph = new TH2F("hDEtaDPhi_hi_sph", ";#Delta#eta;#Delta#phi [rad]", 
                                200, -3., 3., /* eta */
                                400, -10., 10.  /* phi */);
    hSphericity = new TH1F("hSphericity", ";sphericity;#",
                           300, -0.5, 1.5);
    hEta = new TH1F("hEta", ";#eta;#", 100, -3., 3.);
    hPhi = new TH1F("hPhi", ";#phi [rad];#", 300, -7., 7.);

    return kStOk;
}



Int_t StSphericity::Finish()
{
    UninitRead();

    outFile->Write();
    outFile->Close();

    return kStOk;
}



int StSphericity::FillChain(TChain *aChain, char *aDir,
                            const char *aFilter, int aMaxFiles)
{
    TSystem *gSystem = new TSystem();
    if (!gSystem)
    {
        std::cout << "StSphericity[ERROR]: can't allocate memory for TSystem\n";
        return 0;
    }
    void *lDir = gSystem->OpenDirectory(aDir);
    if (!lDir)
    {
        std::cout << "StSphericity[ERROR]: can't open directory " 
            << aDir << std::endl;
        delete gSystem;
        return 0;
    }

    int lCount = 0;
    char *lFileName;

    while ((lFileName = (char *)gSystem->GetDirEntry(aDir)))
    {
        if (strcmp(lFileName, ".") == 0 || strcmp(lFileName, "..") == 0)
            continue;
        if (strstr(lFileName, aFilter))
        {
            char *lFullName = gSystem->ConcatFileName(aDir, lFileName);
            std::cout << "StSphericity[INFO]: Adding "
                << lFullName << " to the chain" << std::endl;
            aChain->Add(lFullName);
            delete lFullName;
            lCount++;
            if ((aMaxFiles > 0) && (lCount > aMaxFiles))   break;
        }
    }

    std::cout << "StSphericity[INFO]: Added " << lCount 
        << " files to the chain" << std::endl;
    delete gSystem;
    return lCount;
}



int StSphericity::FillChain(TChain *aChain, const char *aFileName,
                            int aMaxFiles)
{
    std::ifstream lInStream(aFileName);

    if (!lInStream.is_open())
    {
        std::cout << "StSphericity[ERROR]: can't open file list: "
            << aFileName << std::endl;
        return 0;
    }
    std::cout << "StSphericity[INFO]: using file list: "
        << aFileName << std::endl;

    int lCount = 0;
    char buf[0xFF];
    for ( ; lInStream.good(); )
    {
        lInStream.getline(buf, 0xFF);
        TString lFileName(buf);
        if (lFileName.Contains("root"))
        {
            aChain->Add(buf);
            lCount++;
            if ((aMaxFiles > 0) && (lCount > aMaxFiles)) break;
        }
    }

    std::cout << "StSphericity[INFO]: Added " << lCount
        << " files to the chain" << std::endl;
    return lCount;
}



int StSphericity::InitRead(string aDir, string aFileName,
                           string aFilter, int aMaxFiles)
{
    eventId = 0;
    chain = new TChain("StFemtoDst", "StFemtoDst");

    std::cout << "StSphericity[INFO]: Call InitRead" << std::endl;

    int lNumFiles = 0;
    if (!aFileName.empty())
    {
        if (strstr(aFileName.c_str(), ".lis") ||
            strstr(aFileName.c_str(), ".list"))
        {
            lNumFiles = FillChain(chain, 
                                  (aDir + aFileName).c_str(), 
                                  aMaxFiles);
        }
        else
        {
            chain->Add((aDir + aFileName).c_str());
            lNumFiles++;
        }
    }
    else
    {
        lNumFiles = FillChain(chain, (char *)aDir.c_str(), // not cool
                              aFilter.c_str(), aMaxFiles);
    }
    chain->SetBranchAddress("StFemtoEvent", &femtoEvent);
    nEvents = (unsigned int)chain->GetEntries();
    return lNumFiles;
}




void StSphericity::UninitRead()
{
    if (chain)     delete chain;
    femtoEvent = 0;
    chain   = 0;
}



void StSphericity::AddTrigId(unsigned int aTrigId)
{
    mTrigIdCol.push_back(aTrigId);
}




int StSphericity::GetNEvents()
{
    if (chain)
        return nEvents;
    else
        return -1;
}




Int_t StSphericity::Make()
{
    if (!nEvents)
    {
        std::cout << "StSphericity[INFO]: no events to read\n";
        return 0;
    }

    femtoEvent->Clear();
    int lBytes = chain->GetEntry(eventId++);

    if (nEvents < eventId)
    {
        std::cout << "StSphericity[INFO]: EOF\n";
        return 0;
    }

    if (!lBytes)
    {
        std::cout << "StSphericity[INFO]: no event\n";
        return 0;
    }

    if (femtoEvent)
    {
        unsigned int lTrigId = femtoEvent->GetTriggerId();

        // check trigger
        if (!mTrigIdCol.empty())
        {
            if (((lTrigId & mTrig) == 0) || ((lTrigId & (!mTrig)) != 0))
                return 0;
        }

        TClonesArray *lTracks = femtoEvent->GetFemtoTracks();

        if (lTracks)
        {
            int lNTracks = lTracks->GetEntries();
            int numOfLeadPart = -1;
            float pOfLeadPart = -999.;
            float sphericity = -999.;
            float pt_full = 0.0;
            float m_a = 0., m_diag = 0., m_b = 0.; // (a       m_diag)
                                                   // (m_diag  b     )
            float D = 0.0;   // discriminant
            float lambda[2]; // eigenvalues
            unsigned int nPairs = 0;
            float nPairsNorm;

            for (int i = 0; i < lNTracks; i++)
            {
                StFemtoTrack *lTrack = (StFemtoTrack *)lTracks->At(i);
                float pt = lTrack->GetPt();
                if (fabs(lTrack->GetEta()) < 1.0 && pt > 0.5)
                {
                    if (pOfLeadPart < pt)
                    {
                        numOfLeadPart = i;
                        pOfLeadPart = pt;
                    }
                    float px = lTrack->GetPx();
                    float py = lTrack->GetPy();

                    m_a += px*px/pt;
                    m_b += py*py/pt;
                    m_diag += px*py/pt;
                    pt_full += pt;
                    nPairs++;
                }
            }
            m_a /= pt_full;
            m_b /= pt_full;
            m_diag /= pt_full;

            if (numOfLeadPart < 0 || nPairs < 2)   return kStOk;

            D = pow(m_a + m_b, 2.) - 4.*(m_a*m_b - m_diag*m_diag);
            if (D >= 0.) // two real lambda
            {
                lambda[0] = 0.5*(m_a + m_b + sqrt(D));
                lambda[1] = 0.5*(m_a + m_b - sqrt(D));
                sphericity = 2./(lambda[0] + lambda[1]);
                if (lambda[0] < lambda[1])
                    sphericity *= lambda[0];
                else
                    sphericity *= lambda[1];
                hSphericity->Fill(sphericity);
            }

            nPairsNorm = (float)nPairs/(float)lNTracks;
            StFemtoTrack *leadTrack = (StFemtoTrack *)lTracks->At(numOfLeadPart);
            for (int i = 0; i < lNTracks; i++)
            {
                if (i == numOfLeadPart)   continue;
                StFemtoTrack *lTrack = (StFemtoTrack *)lTracks->At(i);
                if (lTrack->GetPt() > 0.5)
                {
                    float phi = lTrack->GetP().Phi();
                    float eta = lTrack->GetEta();
                    float dEta = eta - leadTrack->GetEta();
                    float dPhi = phi - leadTrack->GetP().Phi();

                    hEta->Fill(eta);
                    hPhi->Fill(phi);
                    hEtaPhi->Fill(eta, phi);
                    hDEtaDPhi->Fill(dEta, dPhi);
                    if (sphericity >= 0. && sphericity < 0.2)
                        hDEtaDPhi_lo_sph->Fill(dEta, dPhi);
                    if (sphericity > 0.8 && sphericity <= 1.0)
                        hDEtaDPhi_hi_sph->Fill(dEta, dPhi);
                }
            }
        }
    }
    return kStOk;
}



void StSphericity::SetTrigger(unsigned int trig, char mode)
{
    unsigned int trigColSize = mTrigIdCol.size();
    for (unsigned int iTrig = 0; iTrig < trigColSize; iTrig++)
    {
        if (mTrigIdCol[iTrig] == trig)
        {
            unsigned int mask = 1 << iTrig;
            if (!mode)
                mTrig = mTrig & (!mask);
            else
                mTrig = mTrig | mask;
            break;
        }
    }

    //
    // Output trigger mask
    //
    char bitMask[33] = {0};
    for (unsigned int iBit = 0; iBit < 32; iBit++)
    {
        bitMask[31 - iBit] = '0' + ((mTrig & (1 << iBit)) >> iBit);
    }
    std::cout << "StSphericity[INFO]: Trigger mask = "
        << mTrig << " bit = "
        << bitMask << std::endl;
}
