#include <TROOT.h>
#include <TSystem.h>
#include <TChain.h>
#include <TFile.h>
#include <TString.h>
#include <iostream>

//
// To define cut monitors
// uncomments line below
//
#define MONITORS

//
// Constants for analyses
//

#define CHARGE_BINS 2
#define SPH_BINS 10

//
//  Event information
//
const int mPrimVertZbins = 6;                // number of z bins for analysis
const double mPrimVertZ[2] = {-30., 30.};    // redefine values of the vertex Z for each run [cm]
const double mPrimVertR = 2.;                // distance in the XY plane [cm]
const double mPrimVertVpdVz[2] = {-5., 5.};  // VpdZ - TPCZ difference
const double mPrimVertXShift = 0.;
const double mPrimVertYShift = 0.;
const bool mUseZdcCorrection = true;         // valid for AuAu collisions. True for 200GeV only!

//
// Track information
//
const double mChargedKaonMass = 0.493677;
const double mChargedPionMass = 0.13957018;
const int mPosCharge = +1;
const int mNegCharge = -1;
const int mTrackType = 1;                     // primary track
const int mPythiaKaonCode = 321;              // pythia particle code
const int mPythiaPionCode = 211;              // pythia particle code

const int mTrackNHits[2] = {10, 50};          // number of track hits
const double mTrackNHitsFit[2] = {5, 50};    // number of track hits used in fit
const double mTrackDCA[2] = {0., 2.};         // DCA of the track and PV [cm]
const double mTrackASplitVal = 0.;
const double mTrackMom[2] = {0.12, 1.55};        // general momentum of the track [GeV/c]
const double mTrackTransvMom[2] = {0.05, 1.55};  // general transverse momentum of the track [GeV/c]
const double mTrackEta[2] = { -0.5, 0.5};

const short mDetectorSelection = 4;  // 0-TPC, 1-ToF, 2-TPC+ToF, 3-ToF||TPC, 4-TPC&&TOF(P>Pthr)||TPC(P<Pthr)

const double mTrackTpcMom[2] = {0.15, 0.55};  // track momentum for TPC idnetification [GeV/c]
const double mnSigmaElectron[2] = {-1e9, 1e9};
const double mnSigmaPion[2] = {-2., 2.};
const double mnSigmaKaon[2] = {-1e9, 1e9};
const double mnSigmaProton[2] = {-1e9, 1e9};

const double mTrackTofMom[2] = {0.15, 1.45};   // track momentum for ToF identification [GeV/c]
const double mPionTofMsqr[2] = {-0.02, 0.062}; // based on p+p data
const double mKaonTofMsqr[2] = {0.2, 0.32};    // based on p+p data

const double mTrackTnTMom[2] = {0.15, 1.55};
const double mTnTNSigmaElectron[2] = {-1e9, 1e9};
const double mTnTNSigmaPion[2] = {-2., 2.};
const double mTnTNSigmaKaon[2] = {-1e9, 1e9};
const double mTnTNSigmaProton[2] = {-1e9, 1e9};

const double mTTTMomThresh = 0.5; // TPC&&TOF (P>Pthr)||TPC(P<Pthr)

//
// Pair information
//
const double mFmrCut[2] = {-1.1, 0.1};         // not more than 10% of merged rows
const double mSplitLevelCut[2] = {-0.5, 0.6};  // splitting level from paper. StHbtPair::quality()
const double mPairKt[2] = {-1e9., 1e9};        // pair transverse momentum cut
const double mPairQinv[2] = {-1e9., 1e9};      // pair relative momentum
const double mAveSep[2] = {5., 5000.};         // average separation of tracks within TPC

//
// Histograms values
//
const int mPairKtBins = 28;                // bin width 
const double mPairKtVal[2] = {0.1, 1.5};   // 50 MeV/c 
const int mQinvBins = 150;                 // 20 MeV/c^2 bin width
const double mQinvVal[2] = {0., 3.0};      // redefine for each analysis

const Char_t *defaultInFile = "test.list";
const Char_t *defaultOutFile = "oExpPion_pp510_1D.root";



void expPionHBT_pp510_y2011_1D(const Char_t *inFileList = defaultInFile,
                               const Char_t *oFileName  = defaultOutFile)
{
    std::cout << "*****************************" << std::endl
              << "*    expPionHBT_pp510_1D    *" << std::endl
              << "*          Start            *" << std::endl
              << "*****************************" << std::endl << std::endl;

    //
    // Load libraries
    //
    std::cout << "Loading libraries..." << std::endl;
    gROOT->LoadMacro("$STAR/StRoot/StMuDSTMaker/COMMON/macros/loadSharedLibraries.C");
    loadSharedLibraries();
    gSystem->Load("libMinuit");
    gSystem->Load("StRefMultCorr");
    gSystem->Load("StFlowMaker");   // should be placed before HBT
    gSystem->Load("StHbtMaker");
    gSystem->Load("StarClassLibrary");
    gSystem->Load("libgsl");
    gSystem->Load("libgslcblas");
    gSystem->Load("StPicoDstMakerRun12");
    gSystem->Load("StarClassLibrary");
    gSystem->Load("libgsl");
    gSystem->Load("libgslcblas");
    gSystem->Load("libVMC.so");
    gSystem->Load("libStDb_Tables.so");
    gSystem->Load("libgen_Tables.so");
    gSystem->Load("libgeometry_Tables.so");
    gSystem->Load("libsim_Tables.so");
    gSystem->Load("libStarMagField.so");
    gSystem->Load("libSt_db_Maker.so");
    gSystem->Load("libSt_g2t.so");
    gSystem->Load("libSt_geant_Maker.so");
    gSystem->Load("StarGeneratorUtil.so");
    gSystem->Load("StarGeneratorEvent.so");
    gSystem->Load("StarGeneratorBase.so");
    gSystem->Load("StFemtoDstMakerSph");

    std::cout << "Libraries have been successfully loaded" << std::endl;

    //
    // Create chain
    //
    StChain* mChain = new StChain("StChain");
    mChain->SetDebug(0);
    StMuDebug::setLevel(0);

    //
    // Set StHbtMaker and StHbtManager
    //
    StHbtMaker *mHbtMaker = new StHbtMaker();
    mHbtMaker->SetDebug(0);
    StHbtManager *mHbtManager = mHbtMaker->HbtManager();

    //
    // Set StHbtReader
    //
    StHbtFemtoDstReader *femtoReader = new StHbtFemtoDstReader("", inFileList, "root", -1, false);
    femtoReader->AddTrigId(320000);   //Trigger selection for pp500 y2011
    femtoReader->AddTrigId(320001);
    femtoReader->AddTrigId(320011);
    femtoReader->AddTrigId(320021);
    femtoReader->AddTrigId(330021);

    femtoReader->AddTrigId(320103);  // NOT VPDMB here and below
    femtoReader->AddTrigId(320113);
    femtoReader->AddTrigId(320123);
    femtoReader->AddTrigId(330123);
    femtoReader->AddTrigId(320500);
    femtoReader->AddTrigId(320504);
    femtoReader->AddTrigId(320514);
    femtoReader->AddTrigId(320524);
    femtoReader->AddTrigId(330524);
    femtoReader->AddTrigId(320501);
    femtoReader->AddTrigId(330501);
    femtoReader->AddTrigId(320503);
    femtoReader->AddTrigId(330503);

    femtoReader->SetTrigger(320000, 1);   // set only VPDMB triggers
    femtoReader->SetTrigger(320001, 1);
    femtoReader->SetTrigger(320011, 1);
    femtoReader->SetTrigger(320021, 1);
    femtoReader->SetTrigger(330021, 1);

    Long64_t mNEvents = femtoReader->GetNEvents();
    mHbtManager->SetEventReader(femtoReader);

    //
    // Set basic event cut
    //
    gregsEventCut *mBasicEventCut = new gregsEventCut();
    mBasicEventCut->SetCollisionType(0);     //0-pp, 1-AuAu
    mBasicEventCut->SetEventMult(0, 45);
    mBasicEventCut->SetVertZPos(mPrimVertZ[0], mPrimVertZ[1]);
    mBasicEventCut->SetVzVpdVzDiff(mPrimVertVpdVz[0],mPrimVertVpdVz[1]);
    mBasicEventCut->SetVertRPos(mPrimVertR);

    //
    // Set basic kaon single-particle cut
    //
    gregsSimpleTrackCut *mBasicTrackCut = new gregsSimpleTrackCut();
    mBasicTrackCut->SetCharge(mPosCharge);
    mBasicTrackCut->SetMass(mChargedPionMass);
    mBasicTrackCut->SetType(mTrackType);
    mBasicTrackCut->SetNHits(mTrackNHits[0], mTrackNHits[1]);
    mBasicTrackCut->SetNHitsFit(mTrackNHitsFit[0], mTrackNHitsFit[1]);
    mBasicTrackCut->SetAntiSplit(mTrackASplitVal);
    mBasicTrackCut->SetEta(mTrackEta[0], mTrackEta[1]);
    mBasicTrackCut->SetDCA(mTrackDCA[0], mTrackDCA[1]);
    mBasicTrackCut->SetDCAGlobal(mTrackDCA[0], mTrackDCA[1]);
    mBasicTrackCut->SetP(mTrackMom[0], mTrackMom[1]);

    mBasicTrackCut->SetDetectorSelection(mDetectorSelection);

    mBasicTrackCut->SetTpcMomentum(mTrackTpcMom[0], mTrackTpcMom[1]);
    mBasicTrackCut->SetNSigmaElectron(mnSigmaElectron[0],mnSigmaElectron[1]);
    mBasicTrackCut->SetNSigmaPion(mnSigmaPion[0], mnSigmaPion[1]);
    mBasicTrackCut->SetNSigmaKaon(mnSigmaKaon[0], mnSigmaKaon[1]);
    mBasicTrackCut->SetNSigmaProton(mnSigmaProton[0], mnSigmaProton[1]);

    mBasicTrackCut->SetTofMassSqr(mPionTofMsqr[0],mPionTofMsqr[1]);
    mBasicTrackCut->SetTofMomentum(mTrackTofMom[0], mTrackTofMom[1]);

    mBasicTrackCut->SetTnTMomentum(mTrackTnTMom[0], mTrackTnTMom[1]);
    mBasicTrackCut->SetTnTNSigmaElectron(mTnTNSigmaElectron[0], mTnTNSigmaElectron[1]);
    mBasicTrackCut->SetTnTNSigmaPion(mTnTNSigmaPion[0], mTnTNSigmaPion[1]);
    mBasicTrackCut->SetTnTNSigmaKaon(mTnTNSigmaKaon[0], mTnTNSigmaKaon[1]);
    mBasicTrackCut->SetTnTNSigmaProton(mTnTNSigmaProton[0], mTnTNSigmaProton[1]);

    mBasicTrackCut->SetTTTMomentumThresh(mTTTMomThresh);

    //
    // Set basic kaon pair cut
    //
    gregsTrackPairCut *mBasicPairCut = new gregsTrackPairCut();
    mBasicPairCut->SetFracOfMergedRow(mFmrCut[0],mFmrCut[1]);
    mBasicPairCut->SetQuality(mSplitLevelCut[0],mSplitLevelCut[1]);
    mBasicPairCut->SetKt(mPairKt[0],mPairKt[1]);
    mBasicPairCut->SetQinv(mPairQinv[0],mPairQinv[1]);
    mBasicPairCut->SetAverageSeparation(mAveSep[0],mAveSep[1]);

    Int_t mLocalCharge, mMultBins, mMultLow, mMultHi, mEvents2Mix, mCurrentPid;

    //
    // The multiplicity analysis matrix is CHARGE_BINS x MULT_BINS, where
    // the first element corresponds to the particle charge: (pos, neg),
    // and the second one corresponds to the RefMult intervals as listed below:
    // Mult: 0-45, 0-2, 3-4, 5-6, 7-8, 9-10, 11-45
    //
    // StHbtAnalysis, cut, CF and monitor declarations
    //
    StHbtVertexMultAnalysis *mPionMultAnalysis[CHARGE_BINS][SPH_BINS];

    //
    // Cuts
    //
    gregsEventCut       *mPionEventCut[CHARGE_BINS][SPH_BINS];
    gregsSimpleTrackCut *mPionTrackCut[CHARGE_BINS][SPH_BINS];
    gregsTrackPairCut   *mPionPairCut[CHARGE_BINS][SPH_BINS];

    //
    // CorrFctns
    //
    QinvCorrFctnKt                *mPionQinvMixKtCF[CHARGE_BINS][SPH_BINS];
    QinvCorrFctnOpposHemisphereKt *mPionQinvOppKtCF[CHARGE_BINS][SPH_BINS];
    QinvCorrFctnRotationKt        *mPionQinvRotKtCF[CHARGE_BINS][SPH_BINS];

#ifdef MONITORS
    //
    // Cut monitors
    //
    fxtEventCutMonitor *mFxtPionEventCutMonitorPass[CHARGE_BINS][SPH_BINS];
    fxtTrackCutMonitor *mFxtPionTrackCutMonitorPass[CHARGE_BINS][SPH_BINS];
    fxtPairCutMonitor  *mFxtPionPairCutMonitorPass[CHARGE_BINS][SPH_BINS];
    fxtEventCutMonitor *mFxtPionEventCutMonitorFail[CHARGE_BINS][SPH_BINS];
    fxtTrackCutMonitor *mFxtPionTrackCutMonitorFail[CHARGE_BINS][SPH_BINS];
    fxtPairCutMonitor  *mFxtPionPairCutMonitorFail[CHARGE_BINS][SPH_BINS];
#endif

    //
    // Analysis creation
    //
    float sphLo, sphHi;
    mMultLow = 0;
    mMultHi = 50;
    mMultBins = 10;
    mEvents2Mix = 10;
    for(Int_t iSph = 0; iSph < SPH_BINS; iSph++)
    {
        //
        // sphLo < sphericity <= sphHi
        //
        switch (iSph)
        {
            case 0:  sphLo = -0.1;     sphHi = 0.1;  break;
            default: sphLo = iSph*0.1; sphHi = (iSph + 1)*0.1;
        }

        for(Int_t iCharge = 0; iCharge < CHARGE_BINS; iCharge++)
        {
            switch(iCharge)
            {
                case 0:   mLocalCharge = mPosCharge;    break;
                default:  mLocalCharge = mNegCharge;    break;
            };

            //
            // Create event cut
            //
            mPionEventCut[iCharge][iSph] = new gregsEventCut(*mBasicEventCut);
            mPionEventCut[iCharge][iSph]->SetEventMult(mMultLow, mMultHi);
            mPionEventCut[iCharge][iSph]->SetSphericityCut(sphLo, sphHi);
#ifdef MONITORS
            mFxtPionEventCutMonitorPass[iCharge][iSph] =
                new fxtEventCutMonitor("eventPionPass", Form("_%d_%d", iCharge, iSph));
            mFxtPionEventCutMonitorFail[iCharge][iSph] =
                new fxtEventCutMonitor("eventPionFail", Form("_%d_%d", iCharge, iSph));
            mPionEventCut[iCharge][iSph]->AddCutMonitor(mFxtPionEventCutMonitorPass[iCharge][iSph],
                                                        mFxtPionEventCutMonitorFail[iCharge][iSph]);
#endif
            //
            //  Like-sign kaon analysis
            //
            // create analysis
            mPionMultAnalysis[iCharge][iSph] = new StHbtVertexMultAnalysis(mPrimVertZbins,
                                                                           mPrimVertZ[0],
                                                                           mPrimVertZ[1],
                                                                           mMultBins,
                                                                           mMultLow,
                                                                           mMultHi);

            mPionMultAnalysis[iCharge][iSph]->SetEventCut(mPionEventCut[iCharge][iSph]);

            //
            // Create single-track cuts and add them to the analysis
            //
            mPionTrackCut[iCharge][iSph] = new gregsSimpleTrackCut(*mBasicTrackCut);
            mPionTrackCut[iCharge][iSph]->SetCharge(mLocalCharge);
#ifdef MONITORS
            mFxtPionTrackCutMonitorPass[iCharge][iSph] =
                new fxtTrackCutMonitor(Form("_trackPionPass_%d_%d_", iCharge, iSph), mChargedPionMass);
            mFxtPionTrackCutMonitorFail[iCharge][iSph] =
                new fxtTrackCutMonitor(Form("_trackPionFail_%d_%d_", iCharge, iSph), mChargedPionMass);
            mPionTrackCut[iCharge][iSph]->AddCutMonitor(mFxtPionTrackCutMonitorPass[iCharge][iSph],
                                                        mFxtPionTrackCutMonitorFail[iCharge][iSph]);
#endif
            mPionMultAnalysis[iCharge][iSph]->SetFirstParticleCut(mPionTrackCut[iCharge][iSph]);
            mPionMultAnalysis[iCharge][iSph]->SetSecondParticleCut(mPionTrackCut[iCharge][iSph]);

            //
            // Create pair cut
            //
            mPionPairCut[iCharge][iSph] = new gregsTrackPairCut(*mBasicPairCut);
#ifdef MONITORS
            mFxtPionPairCutMonitorPass[iCharge][iSph] =
                new fxtPairCutMonitor(Form("_pairPionPass_%d_%d_", iCharge, iSph));
            mFxtPionPairCutMonitorPass[iCharge][iSph]->SetParticleMass(mChargedPionMass);
            mFxtPionPairCutMonitorFail[iCharge][iSph] =
                new fxtPairCutMonitor(Form("_pairPionFail_%d_%d_", iCharge, iSph));
            mFxtPionPairCutMonitorFail[iCharge][iSph]->SetParticleMass(mChargedPionMass);
            mPionPairCut[iCharge][iSph]->AddCutMonitor(mFxtPionPairCutMonitorPass[iCharge][iSph],
                                                       mFxtPionPairCutMonitorFail[iCharge][iSph]);
#endif
            mPionMultAnalysis[iCharge][iSph]->SetPairCut(mPionPairCut[iCharge][iSph]);

            //
            // Correlation function initialization
            //
            mPionQinvMixKtCF[iCharge][iSph] =
                new QinvCorrFctnKt(Form("hPionQinvMixKt_%d_%d", iCharge, iSph),
                                   mQinvBins, mQinvVal[0], mQinvVal[1],
                                   mPairKtBins, mPairKtVal[0], mPairKtVal[1]);
            mPionQinvOppKtCF[iCharge][iSph] =
                new QinvCorrFctnOpposHemisphereKt(Form("hPionQinvOppKt_%d_%d", iCharge, iSph),
                                                  mQinvBins, mQinvVal[0], mQinvVal[1],
                                                  mPairKtBins, mPairKtVal[0], mPairKtVal[1],
                                                  mChargedPionMass, mPionPairCut[iCharge][iSph]);
            mPionQinvRotKtCF[iCharge][iSph] =
                new QinvCorrFctnRotationKt(Form("hPionQinvRotKt_%d_%d", iCharge, iSph),
                                           mQinvBins, mQinvVal[0], mQinvVal[1],
                                           mPairKtBins, mPairKtVal[0], mPairKtVal[1],
                                           mChargedPionMass, mPionPairCut[iCharge][iSph]);

            mPionMultAnalysis[iCharge][iSph]->AddCorrFctn(mPionQinvMixKtCF[iCharge][iSph]);
            mPionMultAnalysis[iCharge][iSph]->AddCorrFctn(mPionQinvOppKtCF[iCharge][iSph]);
            mPionMultAnalysis[iCharge][iSph]->AddCorrFctn(mPionQinvRotKtCF[iCharge][iSph]);

            mPionMultAnalysis[iCharge][iSph]->SetNumEventsToMix(mEvents2Mix); // set number of events to mix
            mHbtManager->AddAnalysis(mPionMultAnalysis[iCharge][iSph]); // add analysis to the manager
        }
    }

    mChain->Init(); // StChain initialization

    //
    // Start processing
    //
    Int_t iRet = 0;
    Int_t mNEventsProcessed = 0;
    Float_t mPercentCounter = 0.05;
    Float_t mProgress = 0.;

    unsigned int mSubCounter = 1;              // time
    time_t mStartTime, mStopTime, mDiffTime;   // time
    float mFrac;                               // time
    mStartTime = time(0);                      // time

    // mNEvents = 15;
    std::cout << "Total Events: " << mNEvents << std::endl;
    for(Int_t iEvent=0; iEvent < mNEvents; iEvent++)
    {

        mProgress = (Float_t)iEvent/(Float_t)mNEvents;
        mNEventsProcessed++;

        if(mProgress >= mPercentCounter)
        {
            mPercentCounter += 0.05;
            mStopTime = time(0);
            mDiffTime = difftime(mStopTime, mStartTime);
            mFrac = (float)mDiffTime*(float)(mNEvents - iEvent)/(float)iEvent;
            mSubCounter++;
            std::cout << Form("Processing progress: %4.2f%%. Time left: %.1f sec", mProgress*100., mFrac)
                << std::endl;
        }

        mChain->Clear();
        iRet = mChain->Make(iEvent);
        if(iRet!=0)
        {
            std::cout << "[ERROR] - iRet: " << iRet << std::endl;
            break;
        }
    }

    mChain->Finish();

    TFile *oFile = new TFile(oFileName, "RECREATE"); // create output file

    //
    // Write hbt histograms to the file
    //
    std::cout << "Writing histograms to the output file...";

#ifdef MONITORS
    for(Int_t iCharge = 0; iCharge < CHARGE_BINS; iCharge++)
    {
        for (Int_t iSph = 0; iSph < SPH_BINS; iSph++)
        {
            mFxtPionEventCutMonitorPass[iCharge][iSph]->VertexZ()->Write();
            mFxtPionEventCutMonitorPass[iCharge][iSph]->VertexYvsVertexX()->Write();
            mFxtPionEventCutMonitorPass[iCharge][iSph]->RefMult()->Write();
            mFxtPionEventCutMonitorPass[iCharge][iSph]->VpdVzDiff()->Write();
            mFxtPionEventCutMonitorFail[iCharge][iSph]->VertexZ()->Write();
            mFxtPionEventCutMonitorFail[iCharge][iSph]->VertexYvsVertexX()->Write();
            mFxtPionEventCutMonitorFail[iCharge][iSph]->RefMult()->Write();
            mFxtPionEventCutMonitorFail[iCharge][iSph]->VpdVzDiff()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->DCAGlobal()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->Nhits()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->P()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->Pt()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->PtVsNsigmaPion()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->PtVsNsigmaKaon()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->PtVsNsigmaProton()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->PvsDedx()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->Rapidity()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->PseudoRapidity()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->PvsMassSqr()->Write();
            mFxtPionTrackCutMonitorPass[iCharge][iSph]->PvsInvBeta()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->DCAGlobal()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->Nhits()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->P()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->Pt()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->PtVsNsigmaPion()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->PtVsNsigmaKaon()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->PtVsNsigmaProton()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->PvsDedx()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->Rapidity()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->PseudoRapidity()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->PvsMassSqr()->Write();
            mFxtPionTrackCutMonitorFail[iCharge][iSph]->PvsInvBeta()->Write();

            mFxtPionPairCutMonitorPass[iCharge][iSph]->Kt()->Write();
            mFxtPionPairCutMonitorPass[iCharge][iSph]->Pt1Pt2Qinv()->Write();
            mFxtPionPairCutMonitorPass[iCharge][iSph]->FMRvsQinv()->Write();
            mFxtPionPairCutMonitorPass[iCharge][iSph]->SLvsQinv()->Write();
            mFxtPionPairCutMonitorPass[iCharge][iSph]->AveSepVsQinv()->Write();
            mFxtPionPairCutMonitorFail[iCharge][iSph]->Kt()->Write();
            mFxtPionPairCutMonitorFail[iCharge][iSph]->Pt1Pt2Qinv()->Write();
            mFxtPionPairCutMonitorFail[iCharge][iSph]->FMRvsQinv()->Write();
            mFxtPionPairCutMonitorFail[iCharge][iSph]->SLvsQinv()->Write();
            mFxtPionPairCutMonitorFail[iCharge][iSph]->AveSepVsQinv()->Write();
        }
    }

#endif
    for (Int_t iCharge = 0; iCharge < CHARGE_BINS; iCharge++)
    {
        for (Int_t iSph = 0; iSph < SPH_BINS; iSph++)
        {
            for (Int_t iKt = 0; iKt < mPairKtBins; iKt++)
            {
                mPionQinvMixKtCF[iCharge][iSph]->Numerator(iKt)->Write();
                mPionQinvMixKtCF[iCharge][iSph]->Denominator(iKt)->Write();
                mPionQinvOppKtCF[iCharge][iSph]->Numerator(iKt)->Write();
                mPionQinvOppKtCF[iCharge][iSph]->Denominator(iKt)->Write();
                mPionQinvRotKtCF[iCharge][iSph]->Numerator(iKt)->Write();
                mPionQinvRotKtCF[iCharge][iSph]->Denominator(iKt)->Write();
            }
        }
    }

    std::cout << "\t[DONE]" << std::endl;

    //
    // Write the output file and close it
    //
    std::cout << "Saving the output file...";
    oFile->Close();
    std::cout << "\t[DONE]" << std::endl;

    delete mChain;

    std::cout << "********************************" << std::endl
        << "*       expPionHBT_pp500_1D    *" << std::endl
        << "*              Finish          *" << std::endl
        << "********************************" << std::endl << std::endl;
}
