
#include <TROOT.h>
#include <TSystem.h>
#include <TChain.h>
#include <TFile.h>
#include <TString.h>
#include <iostream>

//
// To define cut monitors
// uncomments line below
//
#define MONITORS

//
// Constants for analyses
//

#define MULT_BINS 3

using namespace std;

//
//  Event information
//
const int mPrimVertZbins = 6;                // number of z bins for analysis
const double mPrimVertZ[2] = {-30., 30.};    // redefine values of the vertex Z for each run [cm]
const double mPrimVertR = 2.;                // distance in the XY plane [cm]
const double mPrimVertVpdVz[2] = {-5., 5.};  // VpdZ - TPCZ difference
const double mPrimVertXShift = 0.;
const double mPrimVertYShift = 0.;
const bool mUseZdcCorrection = true;         // valid for AuAu collisions. True for 200GeV only!

//
// Track information
//
const double mChargedKaonMass = 0.493677;
const double mChargedPionMass = 0.13957018;
const int mPosCharge = +1;
const int mNegCharge = -1;
const int mTrackType = 1;                     // primary track
const int mPythiaKaonCode = 321;              // pythia particle code
const int mPythiaPionCode = 211;              // pythia particle code

const int mTrackNHits[2] = {15, 50};          // number of track hits
const double mTrackNHitsFit[2] = {5, 50};    // number of track hits used in fit
const double mTrackDCA[2] = {0., 2.};         // DCA of the track and PV [cm]
const double mTrackASplitVal = 0.;
const double mTrackMom[2] = {0.15, 1.55};        // general momentum of the track [GeV/c]
const double mTrackTransvMom[2] = {0.05, 1.55};  // general transverse momentum of the track [GeV/c]
const double mTrackEta[2] = { -1., 1.};

const short mDetectorSelection = 4;  // 0-TPC, 1-ToF, 2-TPC+ToF, 3-ToF||TPC, 4-TPC&&TOF(P>Pthr)||TPC(P<Pthr)

const double mTrackTpcMom[2] = {0.15, 0.55};  // track momentum for TPC idnetification [GeV/c]
const double mnSigmaElectron[2] = {-1e9, 1e9};
const double mnSigmaPion[2] = {-1e9, 1e9};
const double mnSigmaKaon[2] = {-2., 2.};
const double mnSigmaProton[2] = {-1e9, 1e9};

const double mTrackTofMom[2] = {0.15, 1.45};   // track momentum for ToF identification [GeV/c]
const double mPionTofMsqr[2] = {-0.02, 0.062}; // based on p+p data
const double mKaonTofMsqr[2] = {0.2, 0.32};    // based on p+p data

const double mTrackTnTMom[2] = {0.15, 1.55};
const double mTnTNSigmaElectron[2] = {-1e9, 1e9};
const double mTnTNSigmaPion[2] = {-1e9, 1e9};
const double mTnTNSigmaKaon[2] = {-2., 2.};
const double mTnTNSigmaProton[2] = {-1e9, 1e9};

const double mTTTMomThresh = 0.5; // TPC&&TOF (P>Pthr)||TPC(P<Pthr)

//
// Pair information
//
const double mFmrCut[2] = {-1.1, 1.0};         // not more than 10% of merged rows
const double mSplitLevelCut[2] = {-0.5, 1.0};  // splitting level from paper. StHbtPair::quality()
const double mPairKt[2] = {-1e9., 1e9};        // pair transverse momentum cut
const double mPairQinv[2] = {-1e9., 1e9};      // pair relative momentum
const double mAveSep[2] = {5., 5000.};         // average separation of tracks within TPC

//
// Histograms values
//
const int mPairKtBins = 28;                // bin width 
const double mPairKtVal[2] = {0.1, 1.5};   // 50 MeV/c 
const int mQinvBins = 150;                 // 20 MeV/c^2 bin width
const double mQinvVal[2] = {0., 3.0};      // redefine for each analysis

const Char_t *defaultInFile = "test.list";
const Char_t *defaultOutFile = "oSimKaon_pp200_1D.root";



void simKaonHBT_pp200_FemtoDst_1D_unlike(const Char_t *inFileList = defaultInFile,
                                         const Char_t *oFileName  = defaultOutFile)
{
    cout
        << "*****************************\n"
        << "*    simKaonHBT_pp200_1D    *\n"
        << "*          Start            *\n"
        << "*****************************\n\n";

    //
    // Load libraries
    //
    cout << "Loading libraries...\n";
    gROOT->LoadMacro("$STAR/StRoot/StMuDSTMaker/COMMON/macros/loadSharedLibraries.C");
    loadSharedLibraries();
    gSystem->Load("libMinuit");
    gSystem->Load("StRefMultCorr");
    gSystem->Load("StFlowMaker");   // should be placed before HBT
    gSystem->Load("StHbtMaker");
    gSystem->Load("StarClassLibrary");
    gSystem->Load("libgsl");
    gSystem->Load("libgslcblas");
    gSystem->Load("StPicoDstMakerRun11");
    gSystem->Load("StFemtoDstMaker");

    cout << "Libraries have been successfully loaded\n";

    //
    // Create chain
    //
    StChain* mChain = new StChain("StChain");
    mChain->SetDebug(0);
    StMuDebug::setLevel(0);

    //
    // Set StHbtMaker and StHbtManager
    //
    StHbtMaker *mHbtMaker = new StHbtMaker();
    mHbtMaker->SetDebug(0);
    StHbtManager *mHbtManager = mHbtMaker->HbtManager();

    //
    // Set StHbtReader
    //
    StHbtFemtoDstReader *femtoReader = new StHbtFemtoDstReader("", inFileList, "root", -1, false);
    Long64_t mNEvents = femtoReader->GetNEvents();
    mHbtManager->SetEventReader(femtoReader);

    //
    // Set basic event cut
    //
    gregsEventCut *mBasicEventCut = new gregsEventCut();
    mBasicEventCut->SetCollisionType(0);     //0-pp, 1-AuAu
    mBasicEventCut->SetEventMult(0, 45);
    mBasicEventCut->SetVertZPos(mPrimVertZ[0], mPrimVertZ[1]);
    mBasicEventCut->SetVzVpdVzDiff(mPrimVertVpdVz[0],mPrimVertVpdVz[1]);
    mBasicEventCut->SetVertRPos(mPrimVertR);

    //
    // Set basic kaon single-particle cut
    //
    gregsSimpleTrackCut *mBasicTrackCut = new gregsSimpleTrackCut();
    mBasicTrackCut->SetCharge(mPosCharge);
    mBasicTrackCut->SetMass(mChargedKaonMass);
    mBasicTrackCut->SetType(mTrackType);
    mBasicTrackCut->SetNHits(mTrackNHits[0], mTrackNHits[1]);
    mBasicTrackCut->SetNHitsFit(mTrackNHitsFit[0], mTrackNHitsFit[1]);
    mBasicTrackCut->SetAntiSplit(mTrackASplitVal);
    mBasicTrackCut->SetEta(mTrackEta[0], mTrackEta[1]);
    mBasicTrackCut->SetDCA(mTrackDCA[0], mTrackDCA[1]);
    mBasicTrackCut->SetDCAGlobal(mTrackDCA[0], mTrackDCA[1]);
    mBasicTrackCut->SetP(mTrackMom[0], mTrackMom[1]);

    mBasicTrackCut->SetDetectorSelection(mDetectorSelection);

    mBasicTrackCut->SetTpcMomentum(mTrackTpcMom[0], mTrackTpcMom[1]);
    mBasicTrackCut->SetNSigmaElectron(mnSigmaElectron[0],mnSigmaElectron[1]);
    mBasicTrackCut->SetNSigmaPion(mnSigmaPion[0], mnSigmaPion[1]);
    mBasicTrackCut->SetNSigmaKaon(mnSigmaKaon[0], mnSigmaKaon[1]);
    mBasicTrackCut->SetNSigmaProton(mnSigmaProton[0], mnSigmaProton[1]);

    mBasicTrackCut->SetTofMassSqr(mKaonTofMsqr[0],mKaonTofMsqr[1]);
    mBasicTrackCut->SetTofMomentum(mTrackTofMom[0], mTrackTofMom[1]);

    mBasicTrackCut->SetTnTMomentum(mTrackTnTMom[0], mTrackTnTMom[1]);
    mBasicTrackCut->SetTnTNSigmaElectron(mTnTNSigmaElectron[0], mTnTNSigmaElectron[1]);
    mBasicTrackCut->SetTnTNSigmaPion(mTnTNSigmaPion[0], mTnTNSigmaPion[1]);
    mBasicTrackCut->SetTnTNSigmaKaon(mTnTNSigmaKaon[0], mTnTNSigmaKaon[1]);
    mBasicTrackCut->SetTnTNSigmaProton(mTnTNSigmaProton[0], mTnTNSigmaProton[1]);

    mBasicTrackCut->SetTTTMomentumThresh(mTTTMomThresh);

    //
    // Set basic kaon pair cut
    //
    gregsTrackPairCut *mBasicPairCut = new gregsTrackPairCut();
    mBasicPairCut->SetFracOfMergedRow(mFmrCut[0],mFmrCut[1]);
    mBasicPairCut->SetQuality(mSplitLevelCut[0],mSplitLevelCut[1]);
    mBasicPairCut->SetKt(mPairKt[0],mPairKt[1]);
    mBasicPairCut->SetQinv(mPairQinv[0],mPairQinv[1]);
    mBasicPairCut->SetAverageSeparation(mAveSep[0],mAveSep[1]);

    Int_t mLocalCharge, mMultBins, mMultLow, mMultHi, mEvents2Mix, mCurrentPid;

    // HBT vertex and multiplicity analysis
    StHbtVertexMultAnalysis *mKaonMultAnalysis[MULT_BINS];

    //
    // Cuts
    //
    gregsEventCut       *mKaonEventCut[MULT_BINS];
    gregsSimpleTrackCut *mKaonTrackCut[2][MULT_BINS]; // 2 for two particles for a pair
    gregsTrackPairCut   *mKaonPairCut[MULT_BINS];

    //
    // CorrFctns
    //
    QinvCorrFctnKt                *mKaonQinvMixKtCF;
    QinvCorrFctnOpposHemisphereKt *mKaonQinvOppKtCF;
    QinvCorrFctnRotationKt        *mKaonQinvRotKtCF;
    QinvCorrFctn                  *mKaonQinvMixMultCF[MULT_BINS];
    QinvCorrFctnOpposHemisphere   *mKaonQinvOppMultCF[MULT_BINS];
    QinvCorrFctnRotation          *mKaonQinvRotMultCF[MULT_BINS];

#ifdef MONITORS
    //
    // Cut monitors
    //
    fxtEventCutMonitor *mFxtKaonEventCutMonitorPass[MULT_BINS];
    fxtTrackCutMonitor *mFxtKaonTrackCutMonitorPass[2][MULT_BINS]; // 2 for two particles for a pair
    fxtPairCutMonitor  *mFxtKaonPairCutMonitorPass[MULT_BINS];
    fxtEventCutMonitor *mFxtKaonEventCutMonitorFail[MULT_BINS];
    fxtTrackCutMonitor *mFxtKaonTrackCutMonitorFail[2][MULT_BINS];
    fxtPairCutMonitor  *mFxtKaonPairCutMonitorFail[MULT_BINS];
#endif

    //
    // Analysis creation
    //
    for(Int_t iMult = 0; iMult < MULT_BINS; iMult++)
    {
        switch(iMult)
        {
            case 1:  mMultLow = 0;  mMultHi = 7;    mMultBins = 1;  mEvents2Mix = 10; break;
            case 2:  mMultLow = 8;  mMultHi = 200;  mMultBins = 1;  mEvents2Mix = 10; break;
            default: mMultLow = 0;  mMultHi = 200;  mMultBins = 3;  mEvents2Mix = 10; break;
        };

        //
        // Create event cut
        //
        mKaonEventCut[iMult] = new gregsEventCut(*mBasicEventCut);
        mKaonEventCut[iMult]->SetEventMult(mMultLow, mMultHi);
#ifdef MONITORS
        mFxtKaonEventCutMonitorPass[iMult] =
            new fxtEventCutMonitor("eventKaonPass", Form("_%d", iMult));
        mFxtKaonEventCutMonitorFail[iMult] =
            new fxtEventCutMonitor("eventKaonFail", Form("_%d", iMult));
        mKaonEventCut[iMult]->AddCutMonitor(mFxtKaonEventCutMonitorPass[iMult], mFxtKaonEventCutMonitorFail[iMult]);
#endif

        //
        //  Like-sign kaon analysis
        //
        // create analysis
        mKaonMultAnalysis[iMult] = new StHbtVertexMultAnalysis(mPrimVertZbins,
                                                               mPrimVertZ[0],
                                                               mPrimVertZ[1],
                                                               mMultBins,
                                                               mMultLow,
                                                               mMultHi);

        mKaonMultAnalysis[iMult]->SetEventCut(mKaonEventCut[iMult]);

        //
        // Create single-track cuts and add them to the analysis
        //
        mKaonTrackCut[0][iMult] = new gregsSimpleTrackCut(*mBasicTrackCut);
        mKaonTrackCut[0][iMult]->SetCharge(mPosCharge);
        mKaonTrackCut[1][iMult] = new gregsSimpleTrackCut(*mBasicTrackCut);
        mKaonTrackCut[1][iMult]->SetCharge(mNegCharge);

#ifdef MONITORS
        // positive track cut monitors
        mFxtKaonTrackCutMonitorPass[0][iMult] =
            new fxtTrackCutMonitor(Form("_trackKaonPassPos_%d_", iMult), mChargedKaonMass);
        mFxtKaonTrackCutMonitorFail[0][iMult] =
            new fxtTrackCutMonitor(Form("_trackKaonFailPos_%d_", iMult), mChargedKaonMass);
        mKaonTrackCut[0][iMult]->AddCutMonitor(mFxtKaonTrackCutMonitorPass[0][iMult],
                                               mFxtKaonTrackCutMonitorFail[0][iMult]);
        // negative track cut monitors
        mFxtKaonTrackCutMonitorPass[1][iMult] =
            new fxtTrackCutMonitor(Form("_trackKaonPassNeg_%d_", iMult), mChargedKaonMass);
        mFxtKaonTrackCutMonitorFail[1][iMult] =
            new fxtTrackCutMonitor(Form("_trackKaonFailNeg_%d_", iMult), mChargedKaonMass);
        mKaonTrackCut[1][iMult]->AddCutMonitor(mFxtKaonTrackCutMonitorPass[1][iMult],
                                               mFxtKaonTrackCutMonitorFail[1][iMult]);
#endif
        mKaonMultAnalysis[iMult]->SetFirstParticleCut(mKaonTrackCut[0][iMult]);
        mKaonMultAnalysis[iMult]->SetSecondParticleCut(mKaonTrackCut[1][iMult]);

        //
        // Create pair cut
        //
        mKaonPairCut[iMult] = new gregsTrackPairCut(*mBasicPairCut);
#ifdef MONITORS
        mFxtKaonPairCutMonitorPass[iMult] =
            new fxtPairCutMonitor(Form("_pairKaonPass_%d_", iMult));
        mFxtKaonPairCutMonitorPass[iMult]->SetParticleMass(mChargedKaonMass);
        mFxtKaonPairCutMonitorFail[iMult] =
            new fxtPairCutMonitor(Form("_pairKaonFail_%d_", iMult));
        mFxtKaonPairCutMonitorFail[iMult]->SetParticleMass(mChargedKaonMass);
        mKaonPairCut[iMult]->AddCutMonitor(mFxtKaonPairCutMonitorPass[iMult],
                                           mFxtKaonPairCutMonitorFail[iMult]);
#endif
        mKaonMultAnalysis[iMult]->SetPairCut(mKaonPairCut[iMult]);

        //
        // Correlation function initialization
        //
        if (iMult == 0)
        {
            mKaonQinvMixKtCF =
                new QinvCorrFctnKt("hKaonQinvMixKt",
                                   mQinvBins, mQinvVal[0], mQinvVal[1],
                                   mPairKtBins, mPairKtVal[0], mPairKtVal[1]);
            mKaonQinvOppKtCF =
                new QinvCorrFctnOpposHemisphereKt("hKaonQinvOppKt",
                                                  mQinvBins, mQinvVal[0], mQinvVal[1],
                                                  mPairKtBins, mPairKtVal[0], mPairKtVal[1],
                                                  mChargedKaonMass, mKaonPairCut[iMult]);
            mKaonQinvRotKtCF =
                new QinvCorrFctnRotationKt("hKaonQinvRotKt",
                                           mQinvBins, mQinvVal[0], mQinvVal[1],
                                           mPairKtBins, mPairKtVal[0], mPairKtVal[1],
                                           mChargedKaonMass, mKaonPairCut[iMult]);

            mKaonMultAnalysis[iMult]->AddCorrFctn(mKaonQinvMixKtCF);
            mKaonMultAnalysis[iMult]->AddCorrFctn(mKaonQinvOppKtCF);
            mKaonMultAnalysis[iMult]->AddCorrFctn(mKaonQinvRotKtCF);
        }
        mKaonQinvMixMultCF[iMult] =
            new QinvCorrFctn(Form("hKaonQinvMixMult_%d", iMult),
                             mQinvBins, mQinvVal[0], mQinvVal[1]);
        mKaonQinvOppMultCF[iMult] =
            new QinvCorrFctnOpposHemisphere(Form("hKaonQinvOppMult_%d", iMult),
                                            mQinvBins, mQinvVal[0], mQinvVal[1],
                                            mChargedKaonMass, mKaonPairCut[iMult]);
        mKaonQinvRotMultCF[iMult] =
            new QinvCorrFctnRotation(Form("hKaonQinvRotMult_%d", iMult),
                                     mQinvBins, mQinvVal[0], mQinvVal[1],
                                     mChargedKaonMass, mKaonPairCut[iMult]);

        mKaonMultAnalysis[iMult]->AddCorrFctn(mKaonQinvMixMultCF[iMult]);
        mKaonMultAnalysis[iMult]->AddCorrFctn(mKaonQinvOppMultCF[iMult]);
        mKaonMultAnalysis[iMult]->AddCorrFctn(mKaonQinvRotMultCF[iMult]);

        mKaonMultAnalysis[iMult]->SetNumEventsToMix(mEvents2Mix); // set number of events to mix
        mHbtManager->AddAnalysis(mKaonMultAnalysis[iMult]); // add analysis to the manager
    }

    mChain->Init(); // StChain initialization

    //
    // Start processing
    //
    Int_t iRet = 0;
    Int_t mNEventsProcessed = 0;
    Float_t mPercentCounter = 0.05;
    Float_t mProgress = 0.;

    unsigned int mSubCounter = 1;              // time
    time_t mStartTime, mStopTime, mDiffTime;   // time
    float mFrac;                               // time
    mStartTime = time(0);                      // time

    // mNEvents = 15;
    cout << "Total Events: " << mNEvents << endl;
    for(Int_t iEvent=0; iEvent < mNEvents; iEvent++)
    {

        mProgress = (Float_t)iEvent/(Float_t)mNEvents;
        mNEventsProcessed++;

        if(mProgress >= mPercentCounter)
        {
            mPercentCounter += 0.05;
            mStopTime = time(0);
            mDiffTime = difftime(mStopTime, mStartTime);
            mFrac = (float)mDiffTime*(float)(mNEvents - iEvent)/(float)iEvent;
            mSubCounter++;
            cout << Form("Processing progress: %4.2f%%. Time left: %.1f sec", mProgress*100., mFrac)
                << endl;
        }

        mChain->Clear();
        iRet = mChain->Make(iEvent);
        if(iRet!=0)
        {
            cout << "[ERROR] - iRet: " << iRet << endl;
            break;
        }
    }

    mChain->Finish();

    TFile *oFile = new TFile(oFileName, "RECREATE"); // create output file

    //
    // Write hbt histograms to the file
    //
    cout << "Writing histograms to the output file...";

#ifdef MONITORS
    for(Int_t iMult = 0; iMult < MULT_BINS; iMult++)
    {
        //
        // Write event cut monitors
        //
        mFxtKaonEventCutMonitorPass[iMult]->VertexZ()->Write();
        mFxtKaonEventCutMonitorPass[iMult]->VertexYvsVertexX()->Write();
        mFxtKaonEventCutMonitorPass[iMult]->RefMult()->Write();
        mFxtKaonEventCutMonitorPass[iMult]->VpdVzDiff()->Write();
        mFxtKaonEventCutMonitorFail[iMult]->VertexZ()->Write();
        mFxtKaonEventCutMonitorFail[iMult]->VertexYvsVertexX()->Write();
        mFxtKaonEventCutMonitorFail[iMult]->RefMult()->Write();
        mFxtKaonEventCutMonitorFail[iMult]->VpdVzDiff()->Write();
        //
        // Write pair cut monitors
        //
        mFxtKaonPairCutMonitorPass[iMult]->Kt()->Write();
        mFxtKaonPairCutMonitorPass[iMult]->Pt1Pt2Qinv()->Write();
        mFxtKaonPairCutMonitorPass[iMult]->FMRvsQinv()->Write();
        mFxtKaonPairCutMonitorPass[iMult]->SLvsQinv()->Write();
        mFxtKaonPairCutMonitorPass[iMult]->AveSepVsQinv()->Write();
        mFxtKaonPairCutMonitorFail[iMult]->Kt()->Write();
        mFxtKaonPairCutMonitorFail[iMult]->Pt1Pt2Qinv()->Write();
        mFxtKaonPairCutMonitorFail[iMult]->FMRvsQinv()->Write();
        mFxtKaonPairCutMonitorFail[iMult]->SLvsQinv()->Write();
        mFxtKaonPairCutMonitorFail[iMult]->AveSepVsQinv()->Write();

        //
        // Write track cut monitors
        //
        for(Int_t iCharge = 0; iCharge < 2; iCharge++)
        {
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->DCAGlobal()->Write();
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->Nhits()->Write();
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->P()->Write();
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->Pt()->Write();
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->PtVsNsigmaPion()->Write();
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->PtVsNsigmaKaon()->Write();
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->PtVsNsigmaProton()->Write();
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->PvsDedx()->Write();
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->Rapidity()->Write();
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->PseudoRapidity()->Write();
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->PvsMassSqr()->Write();
            mFxtKaonTrackCutMonitorPass[iCharge][iMult]->PvsInvBeta()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->DCAGlobal()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->Nhits()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->P()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->Pt()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->PtVsNsigmaPion()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->PtVsNsigmaKaon()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->PtVsNsigmaProton()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->PvsDedx()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->Rapidity()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->PseudoRapidity()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->PvsMassSqr()->Write();
            mFxtKaonTrackCutMonitorFail[iCharge][iMult]->PvsInvBeta()->Write();
        }
    }

#endif

    //
    // Write corrlation functions to the output file
    //
    for (Int_t iMult = 0; iMult < MULT_BINS; iMult++)
    {
        if (iMult == 0)
        {
            for (Int_t iKt = 0; iKt < mPairKtBins; iKt++)
            {
                mKaonQinvMixKtCF->Numerator(iKt)->Write();
                mKaonQinvMixKtCF->Denominator(iKt)->Write();
                mKaonQinvOppKtCF->Numerator(iKt)->Write();
                mKaonQinvOppKtCF->Denominator(iKt)->Write();
                mKaonQinvRotKtCF->Numerator(iKt)->Write();
                mKaonQinvRotKtCF->Denominator(iKt)->Write();
            }
        }
        mKaonQinvMixMultCF[iMult]->Numerator()->Write();
        mKaonQinvMixMultCF[iMult]->Denominator()->Write();
        mKaonQinvOppMultCF[iMult]->Numerator()->Write();
        mKaonQinvOppMultCF[iMult]->Denominator()->Write();
        mKaonQinvRotMultCF[iMult]->Numerator()->Write();
        mKaonQinvRotMultCF[iMult]->Denominator()->Write();
    }

    cout << "\t[DONE]\n";

    //
    // Writeout the output file and close it
    //
    cout << "Saving the output file...";
    oFile->Close();
    cout << "\t[DONE]\n";

    delete mChain;

    cout
        << "********************************\n"
        << "*       simKaonHBT_pp200_1D    *\n"
        << "*              Finish          *\n"
        << "********************************\n\n";
}
