#include "DataProcessing.h"
#include "FemtoUtils.h"
#include "HbtFitter.h"
#include <TLegend.h>
#include <TCanvas.h>
#include <TMath.h>

const Double_t MHC     = 0.197326960277;  // GeV * fm
const Double_t MHC_SQR = 0.038937929230;  // GeV^2 *fm^2
const Double_t INVSQRTPI = 0.564189584;   // 1/sqrt(pi)

vector<double> DoFit(TH1* histo, Int_t PartType=4,
                     Int_t SourceFunc=0, Int_t NonFemtoFunc=1,
                     Bool_t DrawFitLegend=false, Bool_t ExcludePoints=false); //Histogram, ParticleType, SourceFuncType, NonFemtoFuncType,DrawLegend

vector<double> DoSeparatedFit(TH1* expHisto, TH1* mcHisto, Int_t PartType=4,
                              Int_t SourceFunc=0, Int_t NonFemtoFunc=1,
                              Bool_t DrawFitLegend=false);



DataProcessing::DataProcessing()
{
    mExpFileName = "~/work/star/data/pp200/expMesonHBT_pp200_2012.root";
    mMcFileName = "~/work/star/data/pp200/mcMesonHbtFemtoDst_pp200_tuneA.root";
    mOutFileName = "~/work/star/soft/star_software/processing/pp/oMeson_pp200_1D.root";

    mQinvBins = 120;
    mQinvLow = 0.;
    mQinvHi = 2.4;

    mIsUnlike = false;

    mNormLo = 0.6;
    mNormHi = 1.0;

    mExpFile = NULL;
    mMcFile = NULL;
    mOutFile = NULL;

    mUseMC = true;
    mDoFit = true;
    mProcInt = true;
    mProcMult = true;
    mProcKt = true;
    mIsKaon = true;

    mMixName = "hKaonQinvMixMult";
    mRotName = "hKaonQinvRotMult";
    mOppName = "hKaonQinvOppMult";
    mMixKtName = "hKaonQinvMixKt";
    mRotKtName = "hKaonQinvRotKt";
    mOppKtName = "hKaonQinvOppKt";

    mDetSelection = 2;

    mExpColor = 2;
    mExpLineWidth = 2;
    mExpMarkerStyle = 20;
    mExpMarkerSize = 1.5;

    mMcColor = 4;
    mMcLineWidth = 2;
    mMcMarkerStyle = 24;
    mMcMarkerSize = 1.5;

    mCorrFctnColor = 1;
    mCorrFctnLineWidth = 2;
    mCorrFctnMarkerStyle = 20;
    mCorrFctnMarkerSize = 1.5;

    mYaxisLo = 0.6;
    mYaxisHi = 1.8;

    //Intervals to get
    mMultVals = 2;
    mKtVals = 4;
    mSphVals = 4;

    //How many bins to read
    mMultBins = 10;
    mKtBins = 28;
    mSphBins = 10;

    mSourceFunc = 0;
    mNonFemtoFunc = 1;
    mDrawLegend = true;
}

//_________________
DataProcessing::~DataProcessing() {

}

//_________________
void DataProcessing::SetExpStyle(Int_t color, Int_t width, Int_t style, Double_t size) {
    mExpColor = color;
    mExpLineWidth = width;
    mExpMarkerStyle = style;
    mExpMarkerSize = size;
}

//_________________
void DataProcessing::SetMcStyle(Int_t color, Int_t width, Int_t style, Double_t size) {
    mMcColor = color;
    mMcLineWidth = width;
    mMcMarkerStyle = style;
    mMcMarkerSize = size;
}

//_________________
void DataProcessing::SetCorrFctnStyle(Int_t color, Int_t width, Int_t style, Double_t size) {
    mCorrFctnColor = color;
    mCorrFctnLineWidth = width;
    mCorrFctnMarkerStyle = style;
    mCorrFctnMarkerSize = size;
}

//_________________
void DataProcessing::DoProcessing() {

    std::cout << "Start processing..." << std::endl;
    OpenFiles();
    ReadHistos();
    ProcessHistos();
    if (mDoFit) {
        FitHistos();
        if (!mUseMC)
            PrintFitUncorrectedParms();
        else
            PrintFitCorrectedParms();
    }
    Finish();
    std::cout << "Processing is finished" << std::endl;
}

//_________________
void DataProcessing::OpenFiles() {

    std::cout << "Opening file with experimental data: " << mExpFileName;
    mExpFile = TFile::Open(mExpFileName);
    std::cout << "\t [DONE]" << std::endl;
    std::cout << "Opening file with simulated data: " << mMcFileName;
    mMcFile = TFile::Open(mMcFileName);
    std::cout << "\t [DONE]" << std::endl;
    std::cout << "Creating output file: " << mOutFileName;
    mOutFile = new TFile(mOutFileName,"recreate");
    std::cout << "\t [DONE]" << std::endl;
}

//_________________
void DataProcessing::ReadHistos() {

    std::cout << "Retrieving histograms from files...";
    std::cout
        << "sphericity bins = " << mSphBins 
        << "\nmultiplicity bins = " << mMultBins
        << "\nkt bins = " << mKtBins << std::endl;

    for (int iCharge = 0; iCharge < 2; iCharge++)
    {
        for (int iSph = 0; iSph < mSphBins; iSph++)
        {
            for (int iMult = 0; iMult < mMultBins; iMult++)
            {
                for(Int_t iKt = 0; iKt < mKtBins; iKt++)
                {
                    if (mIsKaon)
                    {
                        //Kaons
                        TH1F *hExpKaonKtMixNum =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                      mMixKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        TH1F *hExpKaonKtMixDen =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                      mMixKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        hExpKaonQinvKtMixNum.push_back(hExpKaonKtMixNum);
                        hExpKaonQinvKtMixDen.push_back(hExpKaonKtMixDen);

                        TH1F *hExpKaonKtRotNum =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                      mRotKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        TH1F *hExpKaonKtRotDen =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                      mRotKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        hExpKaonQinvKtRotNum.push_back(hExpKaonKtRotNum);
                        hExpKaonQinvKtRotDen.push_back(hExpKaonKtRotDen);

                        TH1F *hExpKaonKtOppNum =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                      mOppKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        TH1F *hExpKaonKtOppDen =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                      mOppKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        hExpKaonQinvKtOppNum.push_back(hExpKaonKtOppNum);
                        hExpKaonQinvKtOppDen.push_back(hExpKaonKtOppDen);
                    }
                    else {
                        //Pions
                        TH1F *hExpPionKtMixNum =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                      mMixKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        TH1F *hExpPionKtMixDen =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                      mMixKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        hExpPionQinvKtMixNum.push_back(hExpPionKtMixNum);
                        hExpPionQinvKtMixDen.push_back(hExpPionKtMixDen);

                        TH1F *hExpPionKtRotNum =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                      mRotKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        TH1F *hExpPionKtRotDen =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                      mRotKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        hExpPionQinvKtRotNum.push_back(hExpPionKtRotNum);
                        hExpPionQinvKtRotDen.push_back(hExpPionKtRotDen);

                        TH1F *hExpPionKtOppNum =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                      mOppKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        TH1F *hExpPionKtOppDen =
                            (TH1F*)mExpFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                      mOppKtName,
                                                      iCharge,
                                                      iMult,
                                                      iSph,
                                                      iKt));
                        hExpPionQinvKtOppNum.push_back(hExpPionKtOppNum);
                        hExpPionQinvKtOppDen.push_back(hExpPionKtOppDen);
                    }
                    if(mUseMC) {
                        if (mIsKaon) {
                            //Kaons
                            TH1F *hMcKaonKtMixNum =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                         mMixKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            TH1F *hMcKaonKtMixDen =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                         mMixKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            hMcKaonQinvKtMixNum.push_back(hMcKaonKtMixNum);
                            hMcKaonQinvKtMixDen.push_back(hMcKaonKtMixDen);

                            TH1F *hMcKaonKtRotNum =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                         mRotKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            TH1F *hMcKaonKtRotDen =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                         mRotKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            hMcKaonQinvKtRotNum.push_back(hMcKaonKtRotNum);
                            hMcKaonQinvKtRotDen.push_back(hMcKaonKtRotDen);

                            TH1F *hMcKaonKtOppNum =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                         mOppKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            TH1F *hMcKaonKtOppDen =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                         mOppKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            hMcKaonQinvKtOppNum.push_back(hMcKaonKtOppNum);
                            hMcKaonQinvKtOppDen.push_back(hMcKaonKtOppDen);
                        }
                        else {
                            //Pions
                            TH1F *hMcPionKtMixNum =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                         mMixKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            TH1F *hMcPionKtMixDen =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                         mMixKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            hMcPionQinvKtMixNum.push_back(hMcPionKtMixNum);
                            hMcPionQinvKtMixDen.push_back(hMcPionKtMixDen);

                            TH1F *hMcPionKtRotNum =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                         mRotKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            TH1F *hMcPionKtRotDen =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                         mRotKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            hMcPionQinvKtRotNum.push_back(hMcPionKtRotNum);
                            hMcPionQinvKtRotDen.push_back(hMcPionKtRotDen);

                            TH1F *hMcPionKtOppNum =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Num_bin_%d",
                                                         mOppKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            TH1F *hMcPionKtOppDen =
                                (TH1F*)mMcFile->Get(Form("%s_%d_%d_%d_Den_bin_%d",
                                                         mOppKtName,
                                                         iCharge,
                                                         iMult,
                                                         iSph,
                                                         iKt));
                            hMcPionQinvKtOppNum.push_back(hMcPionKtOppNum);
                            hMcPionQinvKtOppDen.push_back(hMcPionKtOppDen);
                        }
                    }
                } //for(Int_t iKt=0; iKt<20; iKt++)
            }
        }
    } //for(Int_t iCharge=0; iCharge<2; iCharge++)

    std::cout << "\t [DONE]" << std::endl;
}



void DataProcessing::ProcessHistos()
{
    if (mProcKt)
    {
        std::cout << "Processing kT histograms...";
        ProcessKtHistos();
        std::cout << "\t [DONE]" << std::endl;
    }
}



TH1F* DataProcessing::BuildKtCF(Int_t charge, Int_t ktBinLo, Int_t ktBinHi,
                                Int_t sphBinLo, Int_t sphBinHi,
                                Int_t multBinLo, Int_t multBinHi,
                                Int_t refSample, Bool_t isExpData, Bool_t isKaon)
{
    if(charge<0 || charge>2) {
        std::cout << "DataProcessing::BuildKtCF [WARNING] - Wrong charge: " << charge
            << std::endl;
        return 0;
    }

    if(refSample<0 || refSample>2) {
        std::cout << "DataProcessing::BuildKtCF [WARNING] - Wrong reference sample: " << refSample
            << std::endl;
        return 0;
    }

    if(ktBinLo<0 || ktBinHi>=mKtBins) {
        std::cout << "DataProcessing::BuildKtCF [WARNING] - Wrong kt bin interval: " << ktBinLo << "-" << ktBinHi
            << std::endl;
        return 0;
    }

    if(sphBinLo<0 || sphBinHi>=mSphBins) {
        std::cout << "DataProcessing::BuildsphCF [WARNING] - Wrong sphericity bin interval: " << sphBinLo << "-" << sphBinHi
            << std::endl;
        return 0;
    }

    TH1F *hRatio;
    TString sName = "h";
    TString sParticle;
    TString sExp;
    TString sRefSample;
    TString sKtBinLo, sKtBinHi;
    TString sSphBinLo, sSphBinHi;

    //Define particle type
    if(isKaon)
        sParticle = "Kaon";
    else 
        sParticle = "Pion";

    //Define experiment or MC
    if(isExpData)
        sExp = "Exp";
    else
        sExp = "Mc";

    //Define reference sample
    if(refSample == 0) 
        sRefSample = "Mix";
    else if(refSample == 1)
        sRefSample = "Rot";
    else if(refSample == 2)
        sRefSample = "Opp";

    TH1F *hNumSum = new TH1F(Form("hNum%s%s%s_ch%u_multLo%u_multHi%u_sphLo%u_sphHi%u_ktLo%u_ktHi%u",
                                  sExp.Data(), sParticle.Data(), sRefSample.Data(), charge,
                                  multBinLo, multBinHi,
                                  sphBinLo, sphBinHi,
                                  ktBinLo, ktBinHi),
                             "",mQinvBins, mQinvLow, mQinvHi);
    TH1F *hDenSum = new TH1F(Form("hDen%s%s%s_ch%u_multLo%u_multHi%u_sphLo%u_sphHi%u_ktLo%u_ktHi%u",
                                  sExp.Data(), sParticle.Data(), sRefSample.Data(), charge,
                                  multBinLo, multBinHi,
                                  sphBinLo, sphBinHi,
                                  ktBinLo, ktBinHi),
                             "",mQinvBins, mQinvLow, mQinvHi);

    sName += sExp;
    sName += sParticle;
    sName += sRefSample;
    sName += "KtCF_ch";
    sName += charge;
    sName += "_multLo";
    sName += multBinLo;
    sName += "_multHi";
    sName += multBinHi;
    sName += "_sphLo";
    sName += sphBinLo;
    sName += "_sphHi";
    sName += sphBinHi;
    sName += "_ktLo";
    sName += ktBinLo;
    sName += "_ktHi";
    sName += ktBinHi;


    hRatio = new TH1F(sName.Data(),"",mQinvBins,mQinvLow,mQinvHi);

    if(isExpData)
    {
        if(charge<2)
        {
            for (Int_t iMult = multBinLo; iMult <= multBinHi; iMult++)
            {
                for (Int_t iSph = sphBinLo; iSph <= sphBinHi; iSph++)
                {
                    for(Int_t iKt = ktBinLo; iKt <= ktBinHi; iKt++)
                    {
                        int bin = charge*mKtBins*mSphBins*mMultBins + iMult*mKtBins*mSphBins + iSph*mKtBins + iKt;
                        if(isKaon)
                        {
                            if(refSample == 0)
                            {
                                hNumSum->Add(hExpKaonQinvKtMixNum[bin]);
                                hDenSum->Add(hExpKaonQinvKtMixDen[bin]);
                            }
                            else if(refSample == 1)
                            {
                                hNumSum->Add(hExpKaonQinvKtRotNum[bin]);
                                hDenSum->Add(hExpKaonQinvKtRotDen[bin]);
                            }
                            else if(refSample == 2)
                            {
                                hNumSum->Add(hExpKaonQinvKtOppNum[bin]);
                                hDenSum->Add(hExpKaonQinvKtOppDen[bin]);
                            }
                        } //if(isKaon)
                        else
                        {
                            if(refSample == 0)
                            {
                                hNumSum->Add(hExpPionQinvKtMixNum[bin]);
                                hDenSum->Add(hExpPionQinvKtMixDen[bin]);
                            }
                            else if(refSample == 1)
                            {
                                hNumSum->Add(hExpPionQinvKtRotNum[bin]);
                                hDenSum->Add(hExpPionQinvKtRotDen[bin]);
                            }
                            else if(refSample == 2)
                            {
                                hNumSum->Add(hExpPionQinvKtOppNum[bin]);
                                hDenSum->Add(hExpPionQinvKtOppDen[bin]);
                            }
                        }
                    } //for(Int_t iKt=BinLo; iKt<BinHi; iKt++)
                }
            }
        } //if(charge<2)
        else
        {
            for(Int_t iCharge=0; iCharge<2; iCharge++)
            {
                for (Int_t iMult = multBinLo; iMult <= multBinHi; iMult++)
                {
                    for (Int_t iSph = sphBinLo; iSph <= sphBinHi; iSph++)
                    {
                        for(Int_t iKt = ktBinLo; iKt <= ktBinHi; iKt++)
                        {
                            int bin = iCharge*mKtBins*mSphBins*mMultBins + iMult*mKtBins*mSphBins + iSph*mKtBins + iKt;

                            if(isKaon)
                            {
                                if(refSample == 0)
                                {
                                    hNumSum->Add(hExpKaonQinvKtMixNum[bin]);
                                    hDenSum->Add(hExpKaonQinvKtMixDen[bin]);
                                }
                                else if(refSample == 1)
                                {
                                    hNumSum->Add(hExpKaonQinvKtRotNum[bin]);
                                    hDenSum->Add(hExpKaonQinvKtRotDen[bin]);
                                }
                                else if(refSample == 2)
                                {
                                    hNumSum->Add(hExpKaonQinvKtOppNum[bin]);
                                    hDenSum->Add(hExpKaonQinvKtOppDen[bin]);
                                }
                            } //if(isKaon)
                            else
                            {
                                if(refSample == 0)
                                {
                                    hNumSum->Add(hExpPionQinvKtMixNum[bin]);
                                    hDenSum->Add(hExpPionQinvKtMixDen[bin]);
                                }
                                else if(refSample == 1)
                                {
                                    hNumSum->Add(hExpPionQinvKtRotNum[bin]);
                                    hDenSum->Add(hExpPionQinvKtRotDen[bin]);
                                }
                                else if(refSample == 2)
                                {
                                    hNumSum->Add(hExpPionQinvKtOppNum[bin]);
                                    hDenSum->Add(hExpPionQinvKtOppDen[bin]);
                                }
                            }
                        } //for(Int_t iKt=BinLo; iKt<=BinHi; iKt++)
                    }
                }
            } //for(Int_t iCharge=0; iCharge<2; iCharge++)
        } //if(charge==2)
    } //if(isExpData)
    else
    { //MC data
        if(charge<2)
        {
            for (Int_t iMult = multBinLo; iMult <= multBinHi; iMult++)
            {
                for (Int_t iSph = sphBinLo; iSph <= sphBinHi; iSph++)
                {
                    for(Int_t iKt = ktBinLo; iKt <= ktBinHi; iKt++)
                    {
                        int bin = charge*mKtBins*mSphBins*mMultBins + iMult*mKtBins*mSphBins + iSph*mKtBins + iKt;

                        if(isKaon)
                        {
                            if(refSample == 0)
                            {
                                hNumSum->Add(hMcKaonQinvKtMixNum[bin]);
                                hDenSum->Add(hMcKaonQinvKtMixDen[bin]);
                            }
                            else if(refSample == 1)
                            {
                                hNumSum->Add(hMcKaonQinvKtRotNum[bin]);
                                hDenSum->Add(hMcKaonQinvKtRotDen[bin]);
                            }
                            else if(refSample == 2)
                            {
                                hNumSum->Add(hMcKaonQinvKtOppNum[bin]);
                                hDenSum->Add(hMcKaonQinvKtOppDen[bin]);
                            }
                        } //if(isKaon)
                        else
                        {
                            if(refSample == 0)
                            {
                                hNumSum->Add(hMcPionQinvKtMixNum[bin]);
                                hDenSum->Add(hMcPionQinvKtMixDen[bin]);
                            }
                            else if(refSample == 1)
                            {
                                hNumSum->Add(hMcPionQinvKtRotNum[bin]);
                                hDenSum->Add(hMcPionQinvKtRotDen[bin]);
                            }
                            else if(refSample == 2)
                            {
                                hNumSum->Add(hMcPionQinvKtOppNum[bin]);
                                hDenSum->Add(hMcPionQinvKtOppDen[bin]);
                            }
                        }
                    }
                }
            }
        } //if(charge<2)
        else
        {
            for(Int_t iCharge=0; iCharge<2; iCharge++)
            {
                for (Int_t iMult = multBinLo; iMult <= multBinHi; iMult++)
                {
                    for (Int_t iSph = sphBinLo; iSph <= sphBinHi; iSph++)
                    {
                        for(Int_t iKt = ktBinLo; iKt <= ktBinHi; iKt++)
                        {
                            int bin = iCharge*mKtBins*mSphBins*mMultBins + iMult*mKtBins*mSphBins + iSph*mKtBins + iKt;
                            if(isKaon)
                            {
                                if(refSample == 0)
                                {
                                    hNumSum->Add(hMcKaonQinvKtMixNum[bin]);
                                    hDenSum->Add(hMcKaonQinvKtMixDen[bin]);
                                }
                                else if(refSample == 1)
                                {
                                    hNumSum->Add(hMcKaonQinvKtRotNum[bin]);
                                    hDenSum->Add(hMcKaonQinvKtRotDen[bin]);
                                }
                                else if(refSample == 2)
                                {
                                    hNumSum->Add(hMcKaonQinvKtOppNum[bin]);
                                    hDenSum->Add(hMcKaonQinvKtOppDen[bin]);
                                }
                            } //if(isKaon)
                            else
                            {
                                if(refSample == 0)
                                {
                                    hNumSum->Add(hMcPionQinvKtMixNum[bin]);
                                    hDenSum->Add(hMcPionQinvKtMixDen[bin]);
                                }
                                else if(refSample == 1)
                                {
                                    hNumSum->Add(hMcPionQinvKtRotNum[bin]);
                                    hDenSum->Add(hMcPionQinvKtRotDen[bin]);
                                }
                                else if(refSample == 2)
                                {
                                    hNumSum->Add(hMcPionQinvKtOppNum[bin]);
                                    hDenSum->Add(hMcPionQinvKtOppDen[bin]);
                                }
                            }
                        } //for(Int_t iKt=BinLo; iKt<=BinHi; iKt++)
                    }
                }
            }
        } //if(charge==2)
    } //if(!isExpData)

    if (hNumSum->GetEntries() == 0 || hDenSum->GetEntries() == 0)
        std::cout << "FUUUUCK\n";

    Make1DCorrFctn(hRatio,hNumSum,hDenSum, mNormLo, mNormHi);
    std::cout 
        << charge << "_"
        << ktBinLo << "_" << ktBinHi << "_"
        << sphBinLo << "_" << sphBinHi
        << " entries ratio = " << hRatio->GetEntries()
        << " entries num = " << hNumSum->GetEntries() 
        << " entries den = " << hDenSum->GetEntries() << "\n";
//    delete hNumSum;
//    delete hDenSum;

    return hRatio;
}



TH1F* DataProcessing::BuildKtCorrectedCF(Int_t charge, Int_t ktBin, Int_t sphBin, Int_t multBin,
                                         Int_t refSample, Bool_t isKaon)
{
    if(charge<0 || charge>2) {
        std::cout << "DataProcessing::BuildKtCorrectedCF [WARNING] - Wrong charge: " << charge
            << std::endl;
        return 0;
    }

    if(refSample<0 || refSample>2) {
        std::cout << "DataProcessing::BuildKtCorrectedCF [WARNING] - Wrong reference sample: " << refSample
            << std::endl;
        return 0;
    }
    /*
       TH1F *hNumSum = new TH1F("hNumSum",
       "",mQinvBins, mQinvLow, mQinvHi);
       TH1F *hDenSum = new TH1F("hDenSum",
       "",mQinvBins, mQinvLow, mQinvHi);
       */
    TH1F *hRatio;
    TString sName = "hCorr";
    TString sParticle;
    TString sRefSample;

    //Define particle type
    if(isKaon)
        sParticle = "Kaon";
    else
        sParticle = "Pion";

    //Define reference sample
    if(refSample == 0)
        sRefSample = "Mix";
    else if(refSample == 1)
        sRefSample = "Rot";
    else if(refSample == 2)
        sRefSample = "Opp";

    TH1F *hNumSum = new TH1F(Form("hNumCorr%s%s_ch%u_multLo%u_multHi%u_sphLo%u_sphHi%u_ktLo%u_ktHi%u",
                                  sParticle.Data(), sRefSample.Data(), charge,
                                  mVecMultBins[multBin].first, mVecMultBins[multBin].second,
                                  mVecSphBins[sphBin].first, mVecSphBins[sphBin].second,
                                  mVecKtBins[ktBin].first, mVecKtBins[ktBin].second),
                             "",mQinvBins, mQinvLow, mQinvHi);
    TH1F *hDenSum = new TH1F(Form("hMc%s%s_ch%u_multLo%u_multHi%u_sphLo%u_sphHi%u_ktLo%u_ktHi%u",
                                  sParticle.Data(), sRefSample.Data(), charge,
                                  mVecMultBins[multBin].first, mVecMultBins[multBin].second,
                                  mVecSphBins[sphBin].first, mVecSphBins[sphBin].second,
                                  mVecKtBins[ktBin].first, mVecKtBins[ktBin].second),
                             "",mQinvBins, mQinvLow, mQinvHi);

    sName += sParticle;
    sName += sRefSample;
    sName += "KtCF_ch";
    sName += charge;
    sName += "_multLo";
    sName += mVecMultBins[multBin].first;
    sName += "_multHi";
    sName += mVecMultBins[multBin].second;
    sName += "_sphLo";
    sName += mVecSphBins[sphBin].first;
    sName += "_sphHi";
    sName += mVecSphBins[sphBin].second;
    sName += "_ktLo";
    sName += mVecKtBins[ktBin].first;
    sName += "_ktHi";
    sName += mVecKtBins[ktBin].second;

    hRatio = new TH1F(sName.Data(),"",mQinvBins,mQinvLow,mQinvHi);
    int bin = charge*mKtVals*mSphVals*mMultVals + multBin*mSphVals*mKtVals + sphBin*mKtVals + ktBin;

    if(isKaon) {
        if(refSample==0) {
            hNumSum->Add(hExpKaonMixKtCF.at(bin));
            hDenSum->Add(hMcKaonMixKtCF.at(bin));
        }
        else if(refSample==1) {
            hNumSum->Add(hExpKaonRotKtCF.at(bin));
            hDenSum->Add(hMcKaonRotKtCF.at(bin));
        }
        if(refSample==2) {
            hNumSum->Add(hExpKaonOppKtCF.at(bin));
            hDenSum->Add(hMcKaonOppKtCF.at(bin));
        }
    } //if(isKaon)
    else  { //isPion
        if(refSample==0) {
            hNumSum->Add(hExpPionMixKtCF.at(bin));
            hDenSum->Add(hMcPionMixKtCF.at(bin));
        }
        else if(refSample==1) {
            hNumSum->Add(hExpPionRotKtCF.at(bin));
            hDenSum->Add(hMcPionRotKtCF.at(bin));
        }
        if(refSample==2) {
            hNumSum->Add(hExpPionOppKtCF.at(bin));
            hDenSum->Add(hMcPionOppKtCF.at(bin));
        }
    } //isPion

    Correct1DCorrFctn(hRatio,hNumSum,hDenSum);
    delete hNumSum;
//    delete hDenSum;

    return hRatio;
}



void DataProcessing::SetExpHistoStyle(TH1 *h)
{
    h->SetLineColor(mExpColor);
    h->SetLineWidth(mExpLineWidth);
    h->SetMarkerStyle(mExpMarkerStyle);
    h->SetMarkerSize(mExpMarkerSize);
    h->SetMarkerColor(mExpColor);
    h->GetYaxis()->SetRangeUser(mYaxisLo,mYaxisHi);
}



void DataProcessing::SetMcHistoStyle(TH1 *h)
{
    h->SetLineColor(mMcColor);
    h->SetLineWidth(mMcLineWidth);
    h->SetMarkerStyle(mMcMarkerStyle);
    h->SetMarkerSize(mMcMarkerSize);
    h->SetMarkerColor(mMcColor);
    h->GetYaxis()->SetRangeUser(mYaxisLo,mYaxisHi);
}



void DataProcessing::SetHistoStyle(TH1 *h)
{
    h->SetLineColor(mCorrFctnColor);
    h->SetLineWidth(mCorrFctnLineWidth);
    h->SetMarkerStyle(mCorrFctnMarkerStyle);
    h->SetMarkerSize(mCorrFctnMarkerSize);
    h->SetMarkerColor(mCorrFctnColor);
    h->GetYaxis()->SetRangeUser(mYaxisLo,mYaxisHi);
}



void DataProcessing::Finish()
{
    mOutFile->Write();
    mOutFile->Close();
    if (mUseMC)
        mMcFile->Close();
    mExpFile->Close();
}



void DataProcessing::CorrectBins(TH1F *from) {

    TH1F* tmp = new TH1F("tmp","",mQinvBins,mQinvLow,mQinvHi);;

    for(Int_t iBin=0; iBin<(mQinvBins+1); iBin++) {
        tmp->SetBinContent(iBin, from->GetBinContent(iBin));
        tmp->SetBinError(iBin, from->GetBinError(iBin));
    }
    TString newName = "hMc";
    TString name = from->GetName();
    name.Remove(0,1);
    newName += name;

    tmp->SetNameTitle(newName.Data(),from->GetTitle());

    *from = *tmp;
}



void DataProcessing::ProcessKtHistos()
{
    /////////////////////////////////////////
    //               Kt dependence         //
    /////////////////////////////////////////
    Int_t ktBinLo, ktBinHi;
    Int_t sphBinLo, sphBinHi;
    Int_t multBinLo, multBinHi;
    int charge_bins = 3;

    for (int iCharge = 0; iCharge < charge_bins; iCharge++)
    {
        for (Int_t iRefSample=0; iRefSample<3; iRefSample++)
        { //Mix, Rot, Opp
            for (Int_t iMult = 0; iMult < mMultVals; iMult++)
            {
                multBinLo = mVecMultBins[iMult].first;
                multBinHi = mVecMultBins[iMult].second;

                for (Int_t iSph = 0; iSph < mSphVals; iSph++)
                {
                    sphBinLo = mVecSphBins[iSph].first;
                    sphBinHi = mVecSphBins[iSph].second;

                    for(Int_t iKtBin=0; iKtBin<mKtVals; iKtBin++)
                    {
                        ktBinLo = mVecKtBins[iKtBin].first;
                        ktBinHi = mVecKtBins[iKtBin].second;

                        if(mIsKaon == false) { //Pions
                            if(mUseMC) { //Mc
                                if(iRefSample==0) {
                                    hMcPionMixKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                        iRefSample,false,mIsKaon) );
                                    SetMcHistoStyle(hMcPionMixKtCF.back());
                                }
                                else if(iRefSample==1) {
                                    hMcPionRotKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                        iRefSample,false,mIsKaon) );
                                }
                                else if(iRefSample==2) {
                                    hMcPionOppKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                        iRefSample,false,mIsKaon) );
                                }
                            } //if(mUseMC)

                            //Exp
                            if(iRefSample==0) {
                                hExpPionMixKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                     iRefSample,true,mIsKaon) );
                                SetExpHistoStyle(hExpPionMixKtCF.back());
                            }
                            else if(iRefSample==1) {
                                hExpPionRotKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                     iRefSample,true,mIsKaon) );
                            }
                            else if(iRefSample==2) {
                                hExpPionOppKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                     iRefSample,true,mIsKaon) );
                            }
                        }
                        else { //Kaons

                            if(mUseMC) { //Mc
                                if(iRefSample==0) { 
                                    hMcKaonMixKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                        iRefSample,false,mIsKaon) );
                                    SetMcHistoStyle(hMcKaonMixKtCF.back());
                                }
                                else if(iRefSample==1) {
                                    hMcKaonRotKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                        iRefSample,false,mIsKaon) );
                                }
                                else if(iRefSample==2) {
                                    hMcKaonOppKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                        iRefSample,false,mIsKaon) );
                                }
                            } //if(mUseMC)

                            if(iRefSample==0) {
                                hExpKaonMixKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                     iRefSample,true,mIsKaon) );
                                SetExpHistoStyle(hExpKaonMixKtCF.back());
                            }
                            else if(iRefSample==1) {
                                hExpKaonRotKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                     iRefSample,true,mIsKaon) );
                            }
                            else if(iRefSample==2) {
                                hExpKaonOppKtCF.push_back( BuildKtCF(iCharge,ktBinLo,ktBinHi,sphBinLo,sphBinHi,multBinLo,multBinHi,
                                                                     iRefSample,true,mIsKaon) );
                            }
                        } //if(mIsKaon==1)
                    } //for(Int_t iKtBin=0; iKtBin<3; iKtBin++)
                }
            }
        } //for(Int_t iRefSample=0; iRefSample<3; iRefSample++)
    } //for(Int_t iCharge=0; iCharge<3; iCharge++)


    if (mUseMC) {
        for (int iCharge = 0; iCharge < charge_bins; iCharge++)
        {
            for (Int_t iRefSample=0; iRefSample<3; iRefSample++)
            {
                for (Int_t iMult = 0; iMult < mMultVals; iMult++)
                {
                    for (Int_t iSphBin = 0; iSphBin < mSphVals; iSphBin++)
                    {
                        for(Int_t iKtBin=0; iKtBin<mKtVals; iKtBin++)
                        {
                            if(mIsKaon) {
                                if(iRefSample==0) {
                                    hKaonMixKtCF.push_back( BuildKtCorrectedCF(iCharge,iKtBin,iSphBin,iMult,
                                                                               iRefSample,mIsKaon) );
                                    SetHistoStyle(hKaonMixKtCF.back());
                                }
                                else if(iRefSample==1) {
                                    hKaonRotKtCF.push_back( BuildKtCorrectedCF(iCharge,iKtBin,iSphBin,iMult,
                                                                               iRefSample,mIsKaon) );
                                    SetHistoStyle(hKaonRotKtCF.back());
                                }
                                else if(iRefSample==2) {
                                    hKaonOppKtCF.push_back( BuildKtCorrectedCF(iCharge,iKtBin,iSphBin,iMult,
                                                                               iRefSample,mIsKaon) );
                                    SetHistoStyle(hKaonOppKtCF.back());
                                }
                            }
                            else { //isPion
                                if(iRefSample==0) {
                                    hPionMixKtCF.push_back( BuildKtCorrectedCF(iCharge,iKtBin,iSphBin,iMult,
                                                                               iRefSample,mIsKaon) );
                                    SetHistoStyle(hPionMixKtCF.back());
                                }
                                else if(iRefSample==1) {
                                    hPionRotKtCF.push_back( BuildKtCorrectedCF(iCharge,iKtBin,iSphBin,iMult,
                                                                               iRefSample,mIsKaon) );
                                    SetHistoStyle(hPionRotKtCF.back());
                                }
                                else if(iRefSample==2) {
                                    hPionOppKtCF.push_back( BuildKtCorrectedCF(iCharge,iKtBin,iSphBin,iMult,
                                                                               iRefSample,mIsKaon) );
                                    SetHistoStyle(hPionOppKtCF.back());
                                }
                            }
                        } //for(Int_t iKtBin=0; iKtBin<mKtVals; iKtBin++)
                    }
                }
            }
        } //for(Int_t iCharge=0; iCharge<3; iCharge++)  
    }
}

//_________________
void DataProcessing::FitHistos() {

    if (!mUseMC) {
        std::cout << "Fitting uncorrected CFs...";
        FitUncorrectedCF();
        std::cout << "\t [DONE]" << std::endl;
    }
    else {
        std::cout << "Fitting corrected CFs...";
        FitCorrectedCF();
        std::cout << "\t [DONE]" << std::endl;
    }
}

//_________________
void DataProcessing::PrintFitUncorrectedParms() {
    std::cout << "----------- Multiplicity values ----------" << std::endl;
    if (mIsKaon) {
        std::cout << "Kaons:" << std::endl;
        for(Int_t i=0; i<mMultVals; i++) {
            std::cout << Form("R: %3.2f +/- %3.2f ", mKaonMultMixRadii[2*mMultVals+i], 
                              mKaonMultMixRadiiErr[2*mMultVals+i])
                << std::endl;
        }
    }
    else {
        std::cout << "Pions:" << std::endl;
        for(Int_t i=0; i<mMultVals; i++) {
            std::cout << Form("R: %3.2f +/- %3.2f ", mPionMultMixRadii[2*mMultVals+i+1], 
                              mPionMultMixRadiiErr[2*mMultVals+i])
                << std::endl;
        }
    }
}

//_________________
void DataProcessing::PrintFitCorrectedParms() {

    std::cout << "----------- Integrated values ------------" << std::endl;
    std::cout << "Kaons:" << std::endl
        << Form("R: %3.2f +/- %3.2f ", mKaonIntegrMixRadii[0], 
                mKaonIntegrMixRadiiErr[0])
        << std::endl
        << "Pions:" << std::endl
        << Form("R: %3.2f +/- %3.2f ", mPionIntegrMixRadii[0], 
                mPionIntegrMixRadiiErr[0])
        << std::endl << std::endl;


    std::cout << "----------- Multiplicity values ----------" << std::endl;
    std::cout << "Kaons:" << std::endl;
    for(Int_t i=0; i<mMultVals; i++) {
        std::cout << Form("R: %3.2f +/- %3.2f ", mKaonMultMixRadii[2*mMultVals+i], 
                          mKaonMultMixRadiiErr[2*mMultVals+i])
            << std::endl;
    }
    std::cout << "Pions:" << std::endl;
    for(Int_t i=0; i<mMultVals; i++) {
        std::cout << Form("R: %3.2f +/- %3.2f ", mPionMultMixRadii[2*mMultVals+i+1], 
                          mPionMultMixRadiiErr[2*mMultVals+i])
            << std::endl;
    }

    std::cout << "-----------      Kt values     ----------" << std::endl;
    std::cout << "Kaons:" << std::endl;
    for(Int_t i=0; i<mKtVals; i++) {
        std::cout << Form("R: %3.2f +/- %3.2f ", mKaonKtMixRadii[2*mKtVals+i], 
                          mKaonKtMixRadiiErr[2*mKtVals+i])
            << std::endl;
    }
    std::cout << "Pions:" << std::endl;
    for(Int_t i=0; i<mKtVals; i++) {
        std::cout << Form("R: %3.2f +/- %3.2f ", mPionKtMixRadii[2*mKtVals+i], 
                          mPionKtMixRadiiErr[2*mKtVals+i])
            << std::endl;
    }

}

//_________________
void DataProcessing::FitCorrectedCF() {

    std::vector<double> mFitParams;

    ////////////////////////////////////
    //       Integrated values        //
    ////////////////////////////////////

    //Kaons
    if (mIsKaon) {
        for(unsigned int i=1; i<hKaonMixIntegrCF.size(); i++) {
            if(mFitParams.empty()) {
            }
            else {
                mFitParams.clear();
            }
            mFitParams = DoFit(hKaonMixIntegrCF.at(i),4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonIntegrMixRadii.push_back(mFitParams.at(0));
            mKaonIntegrMixRadiiErr.push_back(mFitParams.at(1));
            mKaonIntegrMixLambda.push_back(mFitParams.at(2));
            mKaonIntegrMixLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hKaonRotIntegrCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hKaonRotIntegrCF[i],4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonIntegrRotRadii.push_back(mFitParams.at(0));
            mKaonIntegrRotRadiiErr.push_back(mFitParams.at(1));
            mKaonIntegrRotLambda.push_back(mFitParams.at(2));
            mKaonIntegrRotLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hKaonOppIntegrCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hKaonOppIntegrCF[i],4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonIntegrOppRadii.push_back(mFitParams.at(0));
            mKaonIntegrOppRadiiErr.push_back(mFitParams.at(1));
            mKaonIntegrOppLambda.push_back(mFitParams.at(2));
            mKaonIntegrOppLambdaErr.push_back(mFitParams.at(3));
        }
    } 
    else {
        //Pions
        for(unsigned int i=1; i<hPionMixIntegrCF.size(); i++) {
            if(mFitParams.empty()) {
            }
            else {
                mFitParams.clear();
            }
            mFitParams = DoFit(hPionMixIntegrCF.at(i),3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionIntegrMixRadii.push_back(mFitParams.at(0));
            mPionIntegrMixRadiiErr.push_back(mFitParams.at(1));
            mPionIntegrMixLambda.push_back(mFitParams.at(2));
            mPionIntegrMixLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hPionRotIntegrCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hPionRotIntegrCF[i],3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionIntegrRotRadii.push_back(mFitParams.at(0));
            mPionIntegrRotRadiiErr.push_back(mFitParams.at(1));
            mPionIntegrRotLambda.push_back(mFitParams.at(2));
            mPionIntegrRotLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hPionOppIntegrCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hPionOppIntegrCF[i],3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionIntegrOppRadii.push_back(mFitParams.at(0));
            mPionIntegrOppRadiiErr.push_back(mFitParams.at(1));
            mPionIntegrOppLambda.push_back(mFitParams.at(2));
            mPionIntegrOppLambdaErr.push_back(mFitParams.at(3));
        }
    }

    ////////////////////////////////////
    //       Multiplicity values      //
    ////////////////////////////////////

    if (mIsKaon) { // kaons
        for(unsigned int i=0; i<hKaonMixMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hKaonMixMultCF[i],4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonMultMixRadii.push_back(mFitParams.at(0));
            mKaonMultMixRadiiErr.push_back(mFitParams.at(1));
            mKaonMultMixLambda.push_back(mFitParams.at(2));
            mKaonMultMixLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hKaonRotMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hKaonRotMultCF[i],4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonMultRotRadii.push_back(mFitParams.at(0));
            mKaonMultRotRadiiErr.push_back(mFitParams.at(1));
            mKaonMultRotLambda.push_back(mFitParams.at(2));
            mKaonMultRotLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hKaonOppMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hKaonOppMultCF[i],4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonMultOppRadii.push_back(mFitParams.at(0));
            mKaonMultOppRadiiErr.push_back(mFitParams.at(1));
            mKaonMultOppLambda.push_back(mFitParams.at(2));
            mKaonMultOppLambdaErr.push_back(mFitParams.at(3));
        }
    } 
    else {  // pions
        for(unsigned int i=0; i<hPionMixMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hPionMixMultCF[i],3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionMultMixRadii.push_back(mFitParams.at(0));
            mPionMultMixRadiiErr.push_back(mFitParams.at(1));
            mPionMultMixLambda.push_back(mFitParams.at(2));
            mPionMultMixLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hPionRotMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hPionRotMultCF[i],3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionMultRotRadii.push_back(mFitParams.at(0));
            mPionMultRotRadiiErr.push_back(mFitParams.at(1));
            mPionMultRotLambda.push_back(mFitParams.at(2));
            mPionMultRotLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hPionOppMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hPionOppMultCF[i],3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionMultOppRadii.push_back(mFitParams.at(0));
            mPionMultOppRadiiErr.push_back(mFitParams.at(1));
            mPionMultOppLambda.push_back(mFitParams.at(2));
            mPionMultOppLambdaErr.push_back(mFitParams.at(3));
        }
    }

    ////////////////////////////////////
    //              Kt values         //
    ////////////////////////////////////

    if (mIsKaon) { // kaons
        for(unsigned int i=0; i<hKaonMixKtCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hKaonMixKtCF[i],4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonKtMixRadii.push_back(mFitParams.at(0));
            mKaonKtMixRadiiErr.push_back(mFitParams.at(1));
            mKaonKtMixLambda.push_back(mFitParams.at(2));
            mKaonKtMixLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hKaonRotKtCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hKaonRotKtCF[i],4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonKtRotRadii.push_back(mFitParams.at(0));
            mKaonKtRotRadiiErr.push_back(mFitParams.at(1));
            mKaonKtRotLambda.push_back(mFitParams.at(2));
            mKaonKtRotLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hKaonOppKtCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hKaonOppKtCF[i],4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonKtOppRadii.push_back(mFitParams.at(0));
            mKaonKtOppRadiiErr.push_back(mFitParams.at(1));
            mKaonKtOppLambda.push_back(mFitParams.at(2));
            mKaonKtOppLambdaErr.push_back(mFitParams.at(3));
        }
    }
    else { // pions
        for(unsigned int i=0; i<hPionMixKtCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hPionMixKtCF[i],3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionKtMixRadii.push_back(mFitParams.at(0));
            mPionKtMixRadiiErr.push_back(mFitParams.at(1));
            mPionKtMixLambda.push_back(mFitParams.at(2));
            mPionKtMixLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hPionRotKtCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hPionRotKtCF[i],3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionKtRotRadii.push_back(mFitParams.at(0));
            mPionKtRotRadiiErr.push_back(mFitParams.at(1));
            mPionKtRotLambda.push_back(mFitParams.at(2));
            mPionKtRotLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hPionOppKtCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hPionOppKtCF[i],3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionKtOppRadii.push_back(mFitParams.at(0));
            mPionKtOppRadiiErr.push_back(mFitParams.at(1));
            mPionKtOppLambda.push_back(mFitParams.at(2));
            mPionKtOppLambdaErr.push_back(mFitParams.at(3));
        }
    }

}

//_________________
void DataProcessing::FitUncorrectedCF() {
    std::vector<double> mFitParams;

    ////////////////////////////////////
    //       Multiplicity values      //
    ////////////////////////////////////

    if (mIsKaon) { // kaons
        for(unsigned int i=0; i<hExpKaonMixMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hExpKaonMixMultCF[i],4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonMultMixRadii.push_back(mFitParams.at(0));
            mKaonMultMixRadiiErr.push_back(mFitParams.at(1));
            mKaonMultMixLambda.push_back(mFitParams.at(2));
            mKaonMultMixLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hExpKaonRotMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hExpKaonRotMultCF[i],4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonMultRotRadii.push_back(mFitParams.at(0));
            mKaonMultRotRadiiErr.push_back(mFitParams.at(1));
            mKaonMultRotLambda.push_back(mFitParams.at(2));
            mKaonMultRotLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hExpKaonOppMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hExpKaonOppMultCF[i],4,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mKaonMultOppRadii.push_back(mFitParams.at(0));
            mKaonMultOppRadiiErr.push_back(mFitParams.at(1));
            mKaonMultOppLambda.push_back(mFitParams.at(2));
            mKaonMultOppLambdaErr.push_back(mFitParams.at(3));
        }
    } 
    else {  // pions
        for(unsigned int i=0; i<hExpPionMixMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hExpPionMixMultCF[i],3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionMultMixRadii.push_back(mFitParams.at(0));
            mPionMultMixRadiiErr.push_back(mFitParams.at(1));
            mPionMultMixLambda.push_back(mFitParams.at(2));
            mPionMultMixLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hExpPionRotMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hExpPionRotMultCF[i],3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionMultRotRadii.push_back(mFitParams.at(0));
            mPionMultRotRadiiErr.push_back(mFitParams.at(1));
            mPionMultRotLambda.push_back(mFitParams.at(2));
            mPionMultRotLambdaErr.push_back(mFitParams.at(3));
        }

        for(unsigned int i=0; i<hExpPionOppMultCF.size(); i++) {
            mFitParams.clear();
            mFitParams = DoFit(hExpPionOppMultCF[i],3,mSourceFunc,mNonFemtoFunc,mDrawLegend);
            mPionMultOppRadii.push_back(mFitParams.at(0));
            mPionMultOppRadiiErr.push_back(mFitParams.at(1));
            mPionMultOppLambda.push_back(mFitParams.at(2));
            mPionMultOppLambdaErr.push_back(mFitParams.at(3));
        }
    }



}

//_________________
vector<double> DoFit(TH1* histo, Int_t PartType, Int_t SourceFunc, 
                     Int_t NonFemtoFunc, Bool_t DrawFitLegend, Bool_t mExcludePoints) {

    vector<double> mRetVect;

    Int_t mNbins = histo->GetXaxis()->GetNbins();
    Double_t mFitMinVal = 0.;
    Double_t mDeltaMaxFitVal = -0.005; // -0.005
    Double_t mFitMaxVal = histo->GetXaxis()->GetBinUpEdge(mNbins) + mDeltaMaxFitVal;
    Int_t mFemtoParNum = 2;       //lambda, R
    Int_t mNonFemtoParNum = 0;

    if(NonFemtoFunc == 0)
        mNonFemtoParNum = 1;
    else if(NonFemtoFunc == 1)
        mNonFemtoParNum = 2;
    else if(NonFemtoFunc >= 2 && NonFemtoFunc<=3)
        mNonFemtoParNum = 3;
    else if(NonFemtoFunc >= 4)
        mNonFemtoParNum = 2;

    Int_t mFullParNum = mFemtoParNum + mNonFemtoParNum;
    HbtFitter* mHbtFitter = new HbtFitter(PartType,SourceFunc,NonFemtoFunc);
    mHbtFitter->Set1DMaxVal(histo->GetXaxis()->GetBinUpEdge(mNbins));
    mHbtFitter->Set1DBinsNum(mNbins);
    TF1* mCorrFit = new TF1("mCorrFit",mHbtFitter,&HbtFitter::FitFunction,
                            mFitMinVal,mFitMaxVal,mFullParNum,"HbtFitter","FitFunction");
    mCorrFit->SetParameter(0, 0.5);
    mCorrFit->SetParLimits(0, 0., 1.05);
    mCorrFit->SetParameter(1, 4.);
    mCorrFit->SetParLimits(1, 0., 20.);
    mCorrFit->SetParameter(2, 1.);
    //mCorrFit->SetParLimits(2, 0.9, 1.1);
    mCorrFit->SetLineWidth(3);
    mCorrFit->SetLineColor(kRed);
    //mCorrFit->SetNpx(100);
    //  histo->Fit("mCorrFit","MRQE","ep");
    histo->Fit("mCorrFit","MQE","ep", 0., 0.8);
    mCorrFit->Draw("same");

    if(SourceFunc==0) {
        mRetVect.push_back(mCorrFit->GetParameter(1)*MHC);                 //0 R
        mRetVect.push_back(mCorrFit->GetParError(1)*MHC);                  //1 dR
    }
    else if(SourceFunc==1) {
        mRetVect.push_back(mCorrFit->GetParameter(1)*MHC*INVSQRTPI);       //0 R
        mRetVect.push_back(mCorrFit->GetParError(1)*MHC*INVSQRTPI);        //1 dR
    }
    mRetVect.push_back(mCorrFit->GetParameter(0));                       //2 Lam
    mRetVect.push_back(mCorrFit->GetParError(0));                        //3 dLam
    mRetVect.push_back(mCorrFit->GetChisquare()/mCorrFit->GetNDF());     //4 chi2/ndf
    mRetVect.push_back(mCorrFit->GetParameter(2));                       //5 Norm
    mRetVect.push_back(mCorrFit->GetParError(2));                        //6 dNorm

    if(DrawFitLegend) {
        char temp1[100];
        char* temp2;
        TLegend* mLegend = new TLegend(0.4,0.5,0.9,0.9);
        mLegend->SetFillColor(kWhite);
        mLegend->SetHeader("Fit parameters:");

        mLegend->AddEntry("mCorrFit","Bowler-Sinyukov fit","l");
        sprintf(temp1,"#chi^{2}/n.d.f. = %4.3f / %2d",mCorrFit->GetChisquare(),mCorrFit->GetNDF());
        temp2 = temp1;
        mLegend->AddEntry((TObject*)0,temp2,"");
        sprintf(temp1,"N = %4.3f \\pm %4.3f",mRetVect.at(5),mRetVect.at(6));
        temp2 = temp1;
        mLegend->AddEntry((TObject*)0,temp2,"");
        sprintf(temp1,"#lambda = %4.2f \\pm %4.2f",mRetVect.at(2),mRetVect.at(3));
        temp2 = temp1;
        mLegend->AddEntry((TObject*)0,temp2,"");
        sprintf(temp1,"R = %4.2f \\pm %4.2f",mRetVect.at(0),mRetVect.at(1));
        temp2 = temp1;
        mLegend->AddEntry((TObject*)0,temp2,"");

        mLegend->Draw();
    }
    return mRetVect;
}

//_________________
vector<double> DoSeparatedFit(TH1* hExpHisto, TH1* hMcHisto, 
                              Int_t PartType, Int_t SourceFunc, 
                              Int_t NonFemtoFunc, Bool_t DrawFitLegend) {

    vector<double> mNonFemtoVect;
    vector<double> mRetVect;

    Int_t mNbins = hExpHisto->GetXaxis()->GetNbins();
    Double_t mFitMinVal = 0.;
    Double_t mDeltaMaxFitVal = -0.005; // -0.005
    Double_t mFitMaxVal = hExpHisto->GetXaxis()->GetBinUpEdge(mNbins) + mDeltaMaxFitVal;
    Int_t mFemtoParNum = 2;       //lambda, R
    Int_t mNonFemtoParNum = 0;
    if(NonFemtoFunc == 0)
        mNonFemtoParNum = 1;
    else if(NonFemtoFunc == 1)
        mNonFemtoParNum = 2;
    else if(NonFemtoFunc >= 2 && NonFemtoFunc<=3)
        mNonFemtoParNum = 3;
    else if(NonFemtoFunc >= 4)
        mNonFemtoParNum = 2;

    Int_t mFullParNum = mFemtoParNum + mNonFemtoParNum;
    HbtFitter* mHbtFitter = new HbtFitter(PartType,SourceFunc,NonFemtoFunc);
    mHbtFitter->Set1DMaxVal(hExpHisto->GetXaxis()->GetBinUpEdge(mNbins));
    mHbtFitter->Set1DBinsNum(mNbins);

    TF1* mBckgFunc = new TF1("mBckgFunc",mHbtFitter,&HbtFitter::FitBackgroundFunction,
                             mFitMinVal,mFitMaxVal,mNonFemtoParNum,
                             "HbtFitter","FitBackgroundFunction");
    mBckgFunc->SetParameter(0, 1.);
    mBckgFunc->SetLineWidth(3);
    mBckgFunc->SetLineColor(kBlue);
    mBckgFunc->SetLineStyle(9);
    hMcHisto->Fit("mBckgFunc","MRQE","ep");
    //mBckgFunc->Draw("same");   //I guess this line is not really needed

    for(Int_t i=0; i<mNonFemtoParNum; i++) {
        mNonFemtoVect.push_back(mBckgFunc->GetParameter(i));
    }

    mHbtFitter->Set1DBackgroundParameters(mNonFemtoVect);

    TF1* mSigFunc = new TF1("mSigFunc",mHbtFitter,&HbtFitter::FitSignalFunction,
                            mFitMinVal,mFitMaxVal,(mFemtoParNum+1),
                            "HbtFitter","FitSignalFunction");   //mFemtoParNum+1 = lambda,R + Normal
    mSigFunc->SetParameter(0, 0.5);     //Lambda
    mSigFunc->SetParLimits(0, 0., 1.05);
    mSigFunc->SetParameter(1, 5.);      //R
    mSigFunc->SetParLimits(1, 0., 20.); 
    mSigFunc->SetParameter(2, 1.);      //Normalization
    //mSigFunc->SetParLimits(0, 0., 1.05);
    mSigFunc->SetLineWidth(3);
    mSigFunc->SetLineColor(kRed);
    mSigFunc->SetLineStyle(1);
    //mSigFunc->SetNpx(50);
    hExpHisto->Fit("mSigFunc","MRQE","ep");
    mSigFunc->Draw("same");

    if(SourceFunc==0) {
        mRetVect.push_back(mSigFunc->GetParameter(1)*MHC);            //0 R
        mRetVect.push_back(mSigFunc->GetParError(1)*MHC);             //1 dR
    }
    else if(SourceFunc==1) {
        mRetVect.push_back(mSigFunc->GetParameter(1)*MHC*INVSQRTPI);  //0 R
        mRetVect.push_back(mSigFunc->GetParError(1)*MHC*INVSQRTPI);   //1 dR
    }
    mRetVect.push_back(mSigFunc->GetParameter(0));                  //2 Lam
    mRetVect.push_back(mSigFunc->GetParError(0));                   //3 dLam
    mRetVect.push_back(mSigFunc->GetChisquare()/mSigFunc->GetNDF());//4 chi2/ndf

    if(DrawFitLegend) {
        char temp1[100];
        char* temp2;
        TLegend* mLegend = new TLegend(0.4,0.5,0.9,0.9);
        mLegend->SetFillColor(kWhite);
        mLegend->SetHeader("Fit parameters:");

        mLegend->AddEntry("mSigFunc","Bowler-Sinyukov fit","l");
        sprintf(temp1,"#chi^{2}/n.d.f. = %4.3f/%2d ",mSigFunc->GetChisquare(),mSigFunc->GetNDF());
        temp2 = temp1;
        mLegend->AddEntry((TObject*)0,temp2,"");
        sprintf(temp1,"#lambda = %4.2f \\pm %4.2f",mRetVect.at(2),mRetVect.at(3));
        temp2 = temp1;
        mLegend->AddEntry((TObject*)0,temp2,"");
        sprintf(temp1,"R = %4.2f \\pm %4.2f",mRetVect.at(0),mRetVect.at(1));
        temp2 = temp1;
        mLegend->AddEntry((TObject*)0,temp2,"");
        mLegend->Draw();
    }
    return mRetVect;
}


void DataProcessing::DoRebin(unsigned int val)
{
    if (mIsKaon)
    {
        for (unsigned int i = 0; i < hExpKaonQinvMixNum.size(); i++)
        {
            hExpKaonQinvMixNum[i]->Rebin(val);
            hExpKaonQinvMixDen[i]->Rebin(val);
            hExpKaonQinvRotNum[i]->Rebin(val);
            hExpKaonQinvRotDen[i]->Rebin(val);
            hExpKaonQinvOppNum[i]->Rebin(val);
            hExpKaonQinvOppDen[i]->Rebin(val);
        }
        for (unsigned int i = 0; i < hExpKaonQinvKtMixNum.size(); i++)
        {
            hExpKaonQinvKtMixNum[i]->Rebin(val);
            hExpKaonQinvKtMixDen[i]->Rebin(val);
            hExpKaonQinvKtRotNum[i]->Rebin(val);
            hExpKaonQinvKtRotDen[i]->Rebin(val);
            hExpKaonQinvKtOppNum[i]->Rebin(val);
            hExpKaonQinvKtOppDen[i]->Rebin(val);
        }
        for (unsigned int i = 0; i < hMcKaonQinvMixNum.size(); i++)
        {
            hMcKaonQinvMixNum[i]->Rebin(val);
            hMcKaonQinvMixDen[i]->Rebin(val);
            hMcKaonQinvRotNum[i]->Rebin(val);
            hMcKaonQinvRotDen[i]->Rebin(val);
            hMcKaonQinvOppNum[i]->Rebin(val);
            hMcKaonQinvOppDen[i]->Rebin(val);
        }
        for (unsigned int i = 0; i < hMcKaonQinvKtMixNum.size(); i++)
        {
            hMcKaonQinvKtMixNum[i]->Rebin(val);
            hMcKaonQinvKtMixDen[i]->Rebin(val);
            hMcKaonQinvKtRotNum[i]->Rebin(val);
            hMcKaonQinvKtRotDen[i]->Rebin(val);
            hMcKaonQinvKtOppNum[i]->Rebin(val);
            hMcKaonQinvKtOppDen[i]->Rebin(val);
        }
    }
    else
    {
        for (unsigned int i = 0; i < hExpPionQinvMixNum.size(); i++)
        {
            hExpPionQinvMixNum[i]->Rebin(val);
            hExpPionQinvMixDen[i]->Rebin(val);
            hExpPionQinvRotNum[i]->Rebin(val);
            hExpPionQinvRotDen[i]->Rebin(val);
            hExpPionQinvOppNum[i]->Rebin(val);
            hExpPionQinvOppDen[i]->Rebin(val);
        }
        for (unsigned int i = 0; i < hExpPionQinvKtMixNum.size(); i++)
        {
            hExpPionQinvKtMixNum[i]->Rebin(val);
            hExpPionQinvKtMixDen[i]->Rebin(val);
            hExpPionQinvKtRotNum[i]->Rebin(val);
            hExpPionQinvKtRotDen[i]->Rebin(val);
            hExpPionQinvKtOppNum[i]->Rebin(val);
            hExpPionQinvKtOppDen[i]->Rebin(val);
        }
        for (unsigned int i = 0; i < hMcPionQinvMixNum.size(); i++)
        {
            hMcPionQinvMixNum[i]->Rebin(val);
            hMcPionQinvMixDen[i]->Rebin(val);
            hMcPionQinvRotNum[i]->Rebin(val);
            hMcPionQinvRotDen[i]->Rebin(val);
            hMcPionQinvOppNum[i]->Rebin(val);
            hMcPionQinvOppDen[i]->Rebin(val);
        }
        for (unsigned int i = 0; i < hMcPionQinvKtMixNum.size(); i++)
        {
            hMcPionQinvKtMixNum[i]->Rebin(val);
            hMcPionQinvKtMixDen[i]->Rebin(val);
            hMcPionQinvKtRotNum[i]->Rebin(val);
            hMcPionQinvKtRotDen[i]->Rebin(val);
            hMcPionQinvKtOppNum[i]->Rebin(val);
            hMcPionQinvKtOppDen[i]->Rebin(val);
        }
    }
}



void DataProcessing::AddKtBin(int lo, int hi)
{
    std::pair<int, int> p;
    p.first  = lo;
    p.second = hi;
    mVecKtBins.push_back(p);
}



void DataProcessing::AddMultBin(int lo, int hi)
{
    std::pair<int, int> p;
    p.first  = lo;
    p.second = hi;
    mVecMultBins.push_back(p);
}



void DataProcessing::AddSphBin(int lo, int hi)
{
    std::pair<int, int> p;
    p.first  = lo;
    p.second = hi;
    mVecSphBins.push_back(p);
}
